import pkg_resources
import fsui
from fsui.qt.window import RealWindow
from .application import Application
from .theme import WorkspaceTheme


TEXT_SPACING = 2


class WindowImages:
    def __init__(self):
        self.nw = fsui.Image("pkg://workspace.ui/data/window-shadow-nw.png")
        self.n = fsui.Image("pkg://workspace.ui/data/window-shadow-n.png")
        self.ne = fsui.Image("pkg://workspace.ui/data/window-shadow-ne.png")
        self.e = fsui.Image("pkg://workspace.ui/data/window-shadow-e.png")
        self.se = fsui.Image("pkg://workspace.ui/data/window-shadow-se.png")
        self.s = fsui.Image("pkg://workspace.ui/data/window-shadow-s.png")
        self.sw = fsui.Image("pkg://workspace.ui/data/window-shadow-sw.png")
        self.w = fsui.Image("pkg://workspace.ui/data/window-shadow-w.png")


class WindowBorder(RealWindow):

    _images = None

    @classmethod
    def images(cls):
        if cls._images is None:
            cls._images = WindowImages()
        return cls._images

    def __init__(self, parent, title, child):
        super().__init__(parent, child, title, border=False)
        # self.child = child

        # self.layout = fsui.VerticalLayout(20)
        # self.hori_layout = fsui.HorizontalLayout()
        # self.layout.add(self.hori_layout, fill=True)
        # self.layout.add(child, fill=True, expand=True)

        # self.setWindowFlags(Qt.FramelessWindowHint)
        from fsui.qt import Qt

        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)

    def paintEvent(self, event):
        # noinspection PyNoneFunctionAssignment
        images = self.images()
        w = self.width()
        h = self.height()
        from fsui.qt import QPainter, QPoint, QRect

        painter = QPainter(self)
        painter.drawImage(QPoint(0, 0), images.nw.qimage)
        painter.drawImage(QRect(40, 0, w - 80, 20), images.n.qimage)
        painter.drawImage(QPoint(w - 40, 0), images.ne.qimage)
        painter.drawImage(QRect(w - 20, 40, 20, h - 80), images.e.qimage)
        painter.drawImage(QPoint(w - 40, h - 40), images.se.qimage)
        painter.drawImage(QRect(40, h - 20, w - 80, 20), images.s.qimage)
        painter.drawImage(QPoint(0, h - 40), images.sw.qimage)
        painter.drawImage(QRect(0, 40, 20, h - 80), images.w.qimage)

    def maximize(self, maximized):
        # FIXME: Better to this via event or something
        if maximized:
            self.layout.set_padding(0)
        else:
            self.layout.set_padding(20)
        super().maximize(maximized)

    # def on_close(self):
    #     self.child.on_window_close()
    #     del self.child


class WindowHeader(fsui.Panel):
    def __init__(
        self,
        parent,
        title="",
        menu=False,
        minimizable=True,
        maximizable=True,
        separator=True,
        closable=True,
        background=None,
    ):
        super().__init__(parent)

        if background is None:
            background = (0xFF, 0xFF, 0xFF)
        background_color = fsui.Color(background)
        if background_color.to_hsl().l >= 0.6:
            self.foreground_color = (0x00, 0x00, 0x00)
        else:
            self.foreground_color = (0xFF, 0xFF, 0xFF)
        self.background_color = background_color
        self.set_background_color(background_color)

        self.set_normal_cursor()

        self.layout = fsui.VerticalLayout()
        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        self.menu_button = WindowMenuButton(self)
        hor_layout.add(self.menu_button)
        if not menu:
            self.menu_button.hide()
            hor_layout.add_spacer(20 - TEXT_SPACING)

        self.title_panel = WindowTitlePanel(self)
        hor_layout.add(self.title_panel, expand=True)

        self.minimize_button = WindowMinimizeButton(self)
        hor_layout.add(self.minimize_button)
        if not minimizable:
            self.minimize_button.hide()

        self.maximize_button = WindowMaximizeButton(self)
        hor_layout.add(self.maximize_button)
        if not maximizable:
            self.maximize_button.hide()

        self.close_button = WindowCloseButton(self)
        hor_layout.add(self.close_button)
        if not closable:
            self.close_button.hide()

        if separator:
            self.separator = TitleSeparator(self)
            self.layout.add(self.separator, fill=True)

    def height(self):
        return self.layout.get_min_height()


class Window(fsui.Window):
    def __init__(
        self,
        parent,
        title="",
        minimizable=True,
        maximizable=True,
        menu=False,
        border=True,
        header=True,
        below=False,
        closable=True,
        color=None,
    ):
        if isinstance(parent, Application):
            parent.add_window(self)
            parent = None
        super().__init__(
            parent,
            title,
            minimizable=minimizable,
            maximizable=maximizable,
            native=False,
            separator=False,
            menu=menu,
            border=border,
            header=header,
            below=below,
            closable=closable,
            color=color,
        )
        self.__menu = None

    def add(self, child):
        if hasattr(self, "set_layout"):
            self.set_layout(child)
        else:
            self.layout = child

    def menu(self):
        return self.__menu

    def set_menu(self, menu):
        self.__menu = menu


class Window2(fsui.Panel):
    def __init__(self, parent, title="", maximizable=True):
        if isinstance(parent, Application):
            parent.add_window(self)
            # parent = None
        self.__border = WindowBorder(None, title, self)
        super().__init__(self.__border)
        self.set_background_color(fsui.Color(0xF2, 0xF2, 0xF2))

        self.layout = fsui.VerticalLayout()
        self.__title_layout = fsui.HorizontalLayout()
        self.layout.add(self.__title_layout, fill=True)

        self.__menu_button = WindowMenuButton(self)
        self.__title_layout.add(self.__menu_button)

        self.__title_panel = WindowTitlePanel(self)
        self.__title_layout.add(self.__title_panel, expand=True)

        self.__minimize_button = WindowMinimizeButton(self)
        self.__title_layout.add(self.__minimize_button)

        self._maximize_button = WindowMaximizeButton(self)
        self.__title_layout.add(self._maximize_button)
        if not maximizable:
            print("hiding")
            self._maximize_button.hide()

        self.__close_button = WindowCloseButton(self)
        self.__title_layout.add(self.__close_button)

        self.__content_added = False
        self.__menu = None

    def title(self):
        return self.__border.get_title()

    def minimize(self):
        self.__border.minimize()

    def maximized(self):
        return self.__border.is_maximized()

    def set_maximized(self, maximized):
        self.__border.maximize(maximized)

    def close(self):
        self.__border.close()

    def set_layout(self, layout):
        assert self.__content_added is False
        self.layout.add(layout, fill=True, expand=True)
        self.__content_added = True

    def menu(self):
        return self.__menu

    def set_menu(self, menu):
        self.__menu = menu

    def show(self):
        self.__border.show()

    def position(self):
        return self.__border.get_position()

    def set_position(self, position):
        return self.__border.set_position(position)

    def on_window_close(self):
        pass

    # def set_size(self, size):
    #     self.__border.set_size(size)


class WindowTitlePanel(fsui.Panel):
    def __init__(self, parent):
        super().__init__(parent, paintable=True)
        self.set_background_color(parent.background_color)
        self.set_min_width(40)
        self.set_min_height(40)

        # self.font = fsui.Font("Noto Sans", 15)
        # self.text_color = fsui.Color(0x80, 0x80, 0x80)

        self.pressed = False
        self.window_pos = (-1, -1)
        self.mouse_pos = (-1, -1)

    def on_paint(self):
        theme = WorkspaceTheme.instance()
        dc = self.create_dc()
        dc.set_font(theme.title_font)
        # text = self.parent().title()
        text = self.parent()._window().title()
        tw, th = dc.measure_text(text)
        # FIXME: measurements with Noto Sans is a bit off...
        th += 2
        # # print(self.size()[1], th)
        # if theme.title_glow_color is not None:
        #     dc.set_text_color(theme.title_glow_color)
        #     for dx in [-1, 1]:
        #         for dy in [-1, 1]:
        #             dc.draw_text(text, TEXT_SPACING + dx,
        #                          (self.size()[1] - th) // 2 + dy)
        # dc.set_text_color(theme.title_color)
        dc.set_text_color(self.parent().foreground_color)
        dc.draw_text(text, TEXT_SPACING, (self.size()[1] - th) // 2)

    def on_left_dclick(self):
        # self.parent().maximize(not self.parent().maximized())
        # noinspection PyProtectedMember
        self.parent().maximize_button.on_activate()

    def on_left_down(self):
        self.pressed = True
        window = self.parent().parent()
        self.window_pos = window.position()
        self.mouse_pos = fsui.get_mouse_position()

    def on_left_up(self):
        self.pressed = False

    def on_mouse_motion(self):
        if self.pressed:
            mouse_pos = fsui.get_mouse_position()
            window_pos = (
                self.window_pos[0] + mouse_pos[0] - self.mouse_pos[0],
                self.window_pos[1] + mouse_pos[1] - self.mouse_pos[1],
            )
            window = self.parent().parent()
            window.set_position(window_pos)


class WindowButton(fsui.Panel):
    def __init__(self, parent, name):
        super().__init__(parent, paintable=True)
        self.set_background_color(parent.background_color)
        self.set_min_width(40)
        self.set_min_height(40)
        self.hover = False
        self.pressed = False

        if parent.background_color.to_hsl().l >= 0.6:
            # FIXME: static resources / share between instances
            self.image = fsui.Image(
                "pkg://workspace.ui/data/window-{}.png".format(name)
            )
            self.image_hover = fsui.Image(
                "pkg://workspace.ui/data/window-{}-hover.png".format(name)
            )
            self.image_pressed = fsui.Image(
                "pkg://workspace.ui/data/window-{}-pressed.png".format(name)
            )
        else:
            self.image = fsui.Image(
                "pkg://workspace.ui/data/window-{}-dark.png".format(name)
            )
            self.image_hover = fsui.Image(
                "pkg://workspace.ui/data/window-{}-hover.png".format(name)
            )
            self.image_pressed = fsui.Image(
                "pkg://workspace.ui/data/window-{}-pressed.png".format(name)
            )

    def on_paint(self):
        dc = self.create_dc()
        if self.pressed and self.hover:
            image = self.image_pressed
        elif self.hover:
            image = self.image_hover
        else:
            image = self.image
        dc.draw_image(image, 0, 0)

    def on_activate(self):
        pass

    def on_left_down(self):
        self.pressed = True
        # For menu popups
        self.on_mouse_motion()
        self.refresh()

    def on_mouse_enter(self):
        if not self.hover:
            self.hover = True
            self.refresh()

    def on_mouse_leave(self):
        # Mouse leave event is received when a menu is popped up, so we
        # explicitly check for mouse position.
        self.on_mouse_motion()

    def on_mouse_motion(self):
        new_state = self.is_mouse_over()
        if self.hover != new_state:
            self.hover = new_state
            self.refresh()

    def on_left_up(self):
        self.pressed = False
        self.refresh()
        if self.is_mouse_over():
            self.on_activate()

    def on_left_dclick(self):
        self.on_left_down()


class WindowMenuButton(WindowButton):
    def __init__(self, parent):
        super().__init__(parent, "menu")

    def on_left_down(self):
        super().on_left_down()
        menu = self.window.menu()
        if menu is not None:
            print(menu)
            self.popup_menu(menu, (0, self.size()[1]))
            # Compensate for left up event not being received when menu is
            # closed. Also check is mouse is still hovering.
            self.pressed = False
            self.on_mouse_motion()
            self.refresh()


class WindowMinimizeButton(WindowButton):
    def __init__(self, parent):
        super().__init__(parent, "minimize")

    def on_activate(self):
        window = self.parent().parent()
        window.minimize()
        # FIXME: restore button state when window is is restored


class WindowMaximizeButton(WindowButton):
    def __init__(self, parent):
        super().__init__(parent, "maximize")

    def on_activate(self):
        if self.visible():
            window = self.parent().parent()
            print(window.is_maximized())
            window.set_maximized(not window.is_maximized())
            # FIXME: button is not refreshed when window is maximized
            # self.on_mouse_motion()


class WindowCloseButton(WindowButton):
    def __init__(self, parent):
        super().__init__(parent, "close")

    def on_activate(self):
        window = self.parent().parent()
        window.close()


# FIXME: Rename to TopPanel?
class TitlePanel(fsui.Panel):
    def __init__(self, parent):
        super().__init__(parent)
        self.set_background_color(WorkspaceTheme.instance().title_background)


# FIXME: Rename to TopSeparator?
class TitleSeparator(fsui.Panel):
    def __init__(self, parent):
        super().__init__(parent)
        self.set_background_color(
            WorkspaceTheme.instance().title_separator_color
        )
        self.set_min_height(2)
        self.set_min_width(100)



import fsbc.util
import fstd.mypy
import doctest
import nose.tools


def test_mypy():
    fstd.mypy.check_module(fsbc.util.__name__)


def test_doctest():
    failure_count, test_count = doctest.testmod(fsbc.util)
    nose.tools.assert_equals(failure_count, 0)



import weakref
import traceback
from fsbc.util import unused
from fsui.qt import Qt, QWidget, QPainter
from fsui.qt.helpers import QParent
from fsui.qt.widget import Widget


class Panel(Widget):
    def __init__(self, parent, paintable=False):
        unused(paintable)
        super().__init__(parent)
        self.set_widget(WidgetWithEventHandlers(QParent(parent), self))
        self._widget.move(0, 2000)
        self._widget.setAutoFillBackground(True)

        self._parent = weakref.ref(parent)
        # QWidget.__init__(self, QParent(parent))
        # self.init_widget(parent)
        self.layout = None
        self._painter = None

        # QWidget.__init__(self)
        # self.setParent(parent.get_container())
        # super(Panel, self).__init__(parent.get_container())
        # super().__init__()
        self._ignore_next_left_down_event = False
        # self._widget.setFocusPolicy(Qt.NoFocus)

    def on_left_down(self):
        pass

    def on_left_up(self):
        pass

    def on_left_dclick(self):
        pass

    def on_mouse_motion(self):
        pass

    def on_paint(self):
        pass

    def create_dc(self):
        from .DrawingContext import DrawingContext

        return DrawingContext(self._painter)

    def on_mouse_enter(self):
        pass

    def on_mouse_leave(self):
        pass

    def on_key_press(self, event):
        pass

    def add(self, child):
        if hasattr(self, "set_layout"):
            self.set_layout(child)
        else:
            self.layout = child


# noinspection PyProtectedMember
class WidgetWithEventHandlers(QWidget):
    def __init__(self, parent, owner):
        super().__init__(parent)
        self._owner = owner
        # self._painter = None

    def owner(self):
        return self._owner

    def paintEvent(self, event):
        # if not self._paintable:
        #     #dc = wx.PaintDC(self)
        #     return

        self.owner()._painter = QPainter(self)
        self.owner()._painter.setRenderHint(QPainter.Antialiasing)
        # self._painter.setRenderHint(QPainter.Qt4CompatiblePainting)
        try:
            self.owner().on_paint()
        except Exception:
            traceback.print_exc()
        finally:
            self.owner()._painter = None

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            # A temp code is made in case _ignore_next_left_down_event is
            # altered inside on_left_down.
            ignore = self.owner()._ignore_next_left_down_event
            self.owner()._ignore_next_left_down_event = False
            if ignore:
                print("_ignore_next_left_down_event was True", self)
            else:
                self.owner().on_left_down()

    def keyPressEvent(self, event):
        self.owner().on_key_press(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.owner().on_left_up()

    def mouseMoveEvent(self, event):
        self.owner().on_mouse_motion()

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.owner().on_left_dclick()

    def showEvent(self, event):
        self.owner().on_resize()

    def resizeEvent(self, event):
        self.owner().on_resize()

    def enterEvent(self, _):
        self.owner().on_mouse_enter()

    def leaveEvent(self, _):
        self.owner().on_mouse_leave()



import os

import nose.tools

import fspy.zipfile
import fstd.mypy


def test_mypy():
    fstd.mypy.check_module(fspy.zipfile.__name__)


def test_latin1_name():
    zf = fspy.zipfile.ZipFile(
        os.path.join(os.path.dirname(__file__), "zipfile", "iso-8859-1.zip"),
        "r",
    )
    nose.tools.assert_equals(zf.namelist(), ["F\xf8rde"])
    zf.read("F\xf8rde")
    nose.tools.assert_raises(KeyError, zf.read, "Forde")
    zf.getinfo("F\xf8rde")
    zf.close()


def test_utf8_name():
    zf = fspy.zipfile.ZipFile(
        os.path.join(os.path.dirname(__file__), "zipfile", "utf-8.zip"), "r"
    )
    nose.tools.assert_equals(zf.namelist(), ["F\xf8rde"])
    zf.read("F\xf8rde")
    nose.tools.assert_raises(KeyError, zf.read, "Forde")
    zf.getinfo("F\xf8rde")
    zf.close()



from fsui.qt.qt import init_qt
from fsbc.application import Application as FSBCApplication


class Application(FSBCApplication):
    def __init__(self, name):
        super().__init__(name)
        self.qapplication = init_qt()
        self.on_create()

    def on_create(self):
        pass

    def run(self):
        self.qapplication.exec_()
        self.stop()

    def set_icon(self, icon):
        self.qapplication.setWindowIcon(icon.qicon())

    def run_in_main(self, function, *args, **kwargs):
        from fsui.qt import call_after

        call_after(function, *args, **kwargs)



import weakref

import workspace.path
import workspace.shell
import workspace.ui


ICON_WIDTH = 120
ICON_HEIGHT = 100


class ScrollArea(workspace.ui.Panel):
    """A widget containing the icons in a directory."""

    def __init__(self, parent):
        super().__init__(parent)
        self.selection = None
        self._sort_key_function = None

    def set_sort_key_function(self, function):
        self._sort_key_function = function

    def populate_with_icons(self, path):
        # self.icons.clear()
        items = workspace.path.listdir(path)
        print("populate_with_icons", items)
        x = 0
        y = 0
        width = self.size()[0]
        print("populate_with_icons width =", width)
        for item in sorted(items, key=self._sort_key_function):
            if item.startswith("."):
                continue
            if item.endswith(".fs-info"):
                continue
            if item.endswith(".uaem"):
                continue
            if item.endswith("~"):
                continue
            if item == "Makefile":
                continue
            icon = ShellIcon(self, workspace.path.join(path, item), item)
            icon_width = ICON_WIDTH
            if len(item) >= 45:
                icon_width *= 4
                icon.set_size((icon_width, ICON_HEIGHT))
            elif len(item) >= 30:
                icon_width *= 3
                icon.set_size((icon_width, ICON_HEIGHT))
            elif len(item) > 15:
                icon_width *= 2
                icon.set_size((icon_width, ICON_HEIGHT))
            if x + icon_width > width:
                x = 0
                y += ICON_HEIGHT
            print(item, x, y, "area width", width)
            icon.set_position((x, y))
            x += icon_width


class ShellIcon(workspace.ui.Canvas):
    """A widget representing a file or a directory."""

    STATE_NORMAL = 0
    STATE_SELECTED = 1

    def __init__(self, parent, path, name):
        super().__init__(parent)
        # self.dir_path = dir_path
        self._path = path
        self.name = name
        # self.set_min_size((ICON_WIDTH, ICON_HEIGHT))
        self.set_size((ICON_WIDTH, ICON_HEIGHT))
        # self.set_background_color(workspace.ui.Color(0xff, 0x00, 0xff))
        self._icon = [None, None]
        self._state = self.STATE_NORMAL

    def select(self):
        if self.parent().selection:
            self.parent().selection().set_state(self.STATE_NORMAL)
        self.set_state(self.STATE_SELECTED)
        self.parent().selection = weakref.ref(self)

    def set_state(self, state):
        self._state = state
        self.refresh()

    def path(self):
        return self._path

    #     path = self.dir_path
    #     if not path.endswith(":"):
    #         path += "/"
    #     path += self.name
    #     return path

    def icon(self, state=STATE_NORMAL) -> workspace.ui.Image:
        if self._icon[state] is None:
            is_dir = workspace.path.isdir(self.path())
            image = workspace.shell.icon(self.path(), state=state)
            if image is None and not is_dir:
                image = workspace.shell.default_icon(self.path(), state=state)
            if image is None:
                icon_name = "file"
                if is_dir:
                    icon_name = "folder"
                # if state == self.STATE_NORMAL:
                #     path = "SYS:Icons/FileTypes/{}.fs-info/icon.png".format(
                #             icon_name)
                # else:
                #     path = "SYS:Icons/FileTypes/{}.fs-info/" \
                #             "selected.png".format(icon_name)
                image = workspace.shell.file_type_icon(icon_name, state)
                # image = workspace.ui.Image(path)
            self._icon[state] = image
        return self._icon[state]

    def on_left_down(self):
        self.select()

    def on_left_dclick(self):
        path = self.path()
        name = workspace.path.basename(path)
        app_path = workspace.path.join(path, name + ".py")
        # FIXME: Use ShellOpen for directories too
        if workspace.shell.tool(path) is not None:
            # Icon has associated tool, use ShellOpen even if it is a
            # directory.
            workspace.shell.open(path)
        elif workspace.path.exists(app_path):
            workspace.shell.open(path)
        elif workspace.path.isdir(path):
            # FIXME: Ugly hack (app is set by FileManager.py)
            app.open(path)
        else:
            workspace.shell.open(path)

    def on_paint(self):
        painter = workspace.ui.Painter(self)
        w, h = painter.measure_text(self.name)
        painter.set_text_color(workspace.ui.Color(0x00, 0x00, 0x00, 0x80))
        painter.draw_text(self.name, (self.width() - w) // 2 + 1, 80 + 1)
        painter.set_text_color(workspace.ui.Color(0xFF, 0xFF, 0xFF))
        painter.draw_text(self.name, (self.width() - w) // 2, 80)
        icon = self.icon(self._state)
        # print(icon, icon)
        painter.draw_image(icon, (self.width() - icon.width()) // 2, 10)



import logging
import os
import weakref


class Application:
    def __init__(self, name, singleton=None, file=None):
        self._file = file
        self.name = name
        self.singleton = singleton
        self._windows = []

    def add_window(self, window):
        print("[Application] Adding window", window)
        self._windows.append(weakref.ref(window))

    def add_menu(self, menu):
        print("[Application] Adding menu", menu)

    def run(self):
        from workspace.util.application_runner import ApplicationRunner

        runner = ApplicationRunner(self.name)
        runner.run(self)

    def logger(self):
        logger = logging.getLogger(self.name)
        return logger

    def directory(self, name=""):
        self.logger().debug(self._file)
        dir_path = os.path.dirname(self._file)
        if name:
            dir_path = os.path.join(dir_path, name)
        return dir_path

    def file(self, name=""):
        parts = name.rsplit("/", 1)
        if len(parts) == 1:
            parts.insert(0, "")
        dir_name, file_name = parts
        dir_path = self.directory(dir_name)
        return os.path.join(dir_path, file_name)



import fsui
from fsbc.util import unused
from launcher.ui.skin import Skin


class SettingsHeader(fsui.Group):

    ICON_LEFT = 0
    ICON_RIGHT = 1

    def __init__(
        self, parent, icon, title, subtitle="", icon_position=ICON_RIGHT
    ):
        unused(subtitle)
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        image = icon.image(48)
        self.image_view = fsui.ImageView(self, image)
        if icon_position == self.ICON_LEFT:
            self.layout.add(self.image_view)
            self.layout.add_spacer(20)

        # vert_layout = fsui.VerticalLayout()
        # self.layout.add(
        #     vert_layout, expand=True, fill=False, valign=0.5)
        self.title_label = fsui.HeadingLabel(self, title)
        if Skin.fws() or True:
            font = fsui.Font("Roboto", 26)
            self.title_label.set_font(font)
            self.layout.add(
                self.title_label, expand=True, fill=False, valign=0.0
            )
        else:
            font = self.title_label.get_font()
            font.increase_size(3)
            self.title_label.set_font(font)
            self.layout.add(
                self.title_label, expand=True, fill=False, valign=0.5
            )

        if icon_position == self.ICON_RIGHT:
            self.layout.add_spacer(20)
            self.layout.add(self.image_view)



import os
import tempfile
import fsbc.settings
import nose.tools


# def test_unload():
#     fd, path = tempfile.mkstemp()
#     os.close(fd)
#     fsbc.Settings.set_path(path)
#     fsbc.Settings.load()
#     os.unlink(path)
#
#     fsbc.Settings.set("Key-å", "Value-å")
#     fsbc.Settings.unload()
#
#     fsbc.Settings.set_path("")
#
#     value = fsbc.Settings.get("Key-å")
#     assert_equals(value, "")


def test_load_save():
    fd, path = tempfile.mkstemp()
    os.close(fd)
    fsbc.settings.set_path(path)
    fsbc.settings.load()

    fsbc.settings.set("Key-å", "Value-å")
    fsbc.settings.save()
    fsbc.settings.unload()

    fsbc.settings.load()

    value = fsbc.settings.get("key-å")
    nose.tools.assert_equals(value, "Value-å")
    fsbc.settings.unload()


def test_mypy():
    # from nose.plugins.skip import SkipTest
    # raise SkipTest()
    import fstd.mypy

    fstd.mypy.check_module(fsbc.settings.__name__)


def test_doctest():
    import doctest

    failure_count, test_count = doctest.testmod(fsbc.settings)
    nose.tools.assert_equals(failure_count, 0)



from fsbc.util import unused
from fsui import Panel, Color, VerticalLayout, HorizontalLayout, Image
from ...launcher_config import LauncherConfig
from ..skin import Skin
from .LanguageElement import LanguageElement
from .PlayersElement import PlayersElement
from .ProtectionElement import ProtectionElement
from .WarningsElement import WarningsElement
from .WebLinkElement import WebLinkElement


class StatusBar(Panel):
    def __init__(self, parent):
        Panel.__init__(self, parent, paintable=True)
        self.set_min_height(29)
        # self.set_background_color(Color(0xd8, 0xd8, 0xd8))
        self.layout = VerticalLayout()
        self.hori_layout = HorizontalLayout()
        if Skin.fws():
            self.top_border_size = 2
        else:
            self.top_border_size = 1
        self.layout.add(
            self.hori_layout,
            fill=True,
            expand=True,
            margin_top=self.top_border_size,
        )

        element = ProtectionElement(self)
        self.hori_layout.add(element, fill=True)

        element = WarningsElement(self)
        self.hori_layout.add(element, fill=True, expand=True)
        self.hori_layout.add_spacer(16)

        for language, icon_name in reversed(
            [
                ("en", "flag-gb"),
                ("de", "flag-de"),
                ("fr", "flag-fr"),
                ("es", "flag-es"),
                ("it", "flag-it"),
                ("pl", "flag-pl"),
                ("sv", "flag-se"),
                ("da", "flag-dk"),
                ("nb", "flag-no"),
                ("nn", "flag-no"),
                ("no", "flag-no"),
                ("fi", "flag-fi"),
                ("ja", "flag-jp"),
                # ("", "flag-unknown"),
            ]
        ):
            icon = Image("workspace:res/16x16/" + icon_name + ".png")
            element = LanguageElement(self, language, icon)
            self.hori_layout.add(element, fill=True)
        self.hori_layout.add_spacer(16)

        element = PlayersElement(self)
        self.hori_layout.add(element, fill=True)

        # for config_key, icon_name in [
        #     ("database_url", "database_url_16"),
        #     ("hol_url", "hol_url_16"),
        #     ("lemonamiga_url", "lemon_url_16"),
        #     ("mobygames_url", "mobygames_url_16"),
        #     ("wikipedia_url", "wikipedia_url_16"),
        # ]:
        #     icon = Image("launcher:res/" + icon_name + ".png")
        #     element = LinkButtonElement(self, config_key, icon)
        #     self.hori_layout.add(element)

        element = WebLinkElement(self)
        self.hori_layout.add(element, fill=True)

        # this listener is added after all status bar children have
        # added their listeners, this is important for re-layout...
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        unused(value)
        layout = False
        if key in ["languages", "protection"]:
            layout = True
        if layout:
            self.layout.update()

    def on_paint(self):
        dc = self.create_dc()
        size = self.size()
        if Skin.fws():
            dc.draw_rectangle(0, 0, size[0], 2, Color(0xE5, 0xE5, 0xE5, 0xFF))
            return
        color_1 = Skin.get_background_color()
        if color_1 is not None:
            color_1 = color_1.copy().darken(0.12)
        else:
            color_1 = Color(0x00, 0x00, 0x00, 0x30)
        dc.draw_line(0, 0, size[0], 0, color_1)
        self.draw_background(self, dc, offset=1, height=size[1] - 1)

    @classmethod
    def draw_element_background(cls, widget, dc):
        cls.draw_background(widget, dc)

    @classmethod
    def draw_background(cls, widget, dc, offset=None, height=None):
        size = widget.size()
        if Skin.fws():
            return
        x = 0
        y = 0
        w = size[0]
        h = size[1]
        if offset is not None:
            y += offset
        if height is not None:
            h = height

        color_1 = Skin.get_background_color()
        color_2 = color_1
        # if fsui.System.macosx:
        #     color_1 = Color(0xa7, 0xa7, 0xa7)
        #     color_2 = Color(0xc0, 0xc0, 0xc0)
        if color_1 is not None:
            color_1 = color_1.copy().darken(0.08)
        else:
            color_1 = Color(0x00, 0x00, 0x00, 0x20)
            color_2 = Color(0x00, 0x00, 0x00, 0x00)
        dc.draw_vertical_gradient(x, y, w, h, color_1, color_2)



from __future__ import absolute_import
from __future__ import print_function

from .Block import Block

class FileListBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, is_type=Block.T_LIST, is_sub_type=Block.ST_FILE)
  
  def set(self, data):
    self._set_data(data)
    self._read()
  
  def read(self):
    self._read_data()
    self._read()
  
  def _read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    # FileList fields
    self.own_key = self._get_long(1)
    self.block_count = self._get_long(2)
    
    # read (limited) data blocks
    bc = self.block_count
    mbc = self.blkdev.block_longs - 56
    if bc > mbc:
      bc = mbc
    self.data_blocks = []
    for i in xrange(bc):
      self.data_blocks.append(self._get_long(-51-i))
    
    self.parent = self._get_long(-3)
    self.extension = self._get_long(-2)

    self.valid = (self.own_key == self.blk_num)
    return self.valid
  
  def create(self, parent, data_blocks, extension):
    Block.create(self)
    self.own_key = self.blk_num
    self.block_count = len(data_blocks)
    self.data_blocks = data_blocks
    self.parent = parent
    self.extension = extension
    self.valid  = True
    return True
    
  def write(self):
    Block._create_data(self)
    self._put_long(1, self.own_key)
    self._put_long(2, self.block_count)
    
    # data blocks
    for i in xrange(len(self.data_blocks)):
      self._put_long(-51-i, self.data_blocks[i])
    
    self._put_long(-3, self.parent)
    self._put_long(-2, self.extension)
    Block.write(self)
  
  def dump(self):
    Block.dump(self,"FileList")
    print(" own_key:    %d" % self.own_key)
    print(" blk_cnt:    %d" % self.block_count)
    print(" data blks:  %s" % self.data_blocks)
    print(" parent:     %d" % self.parent)
    print(" extension:  %d" % self.extension)
  


from .Button import Button
from .Panel import Panel
from ..common.i18n import gettext
from ..common.layout import HorizontalLayout, VerticalLayout


class DialogButtons(Panel):
    def __init__(self, parent):
        super().__init__(parent)
        self.layout = HorizontalLayout(padding=20)
        self.layout.padding_top = 0
        self.added_spacer = False
        self.button_layout = HorizontalLayout()
        self.layout.add_spacer(0, expand=True)
        self.layout.add(self.button_layout, fill=True)

    @classmethod
    def create_with_layout(cls, parent):
        parent.layout = VerticalLayout()
        buttons = DialogButtons(parent)
        layout = VerticalLayout(padding=20)
        parent.layout.add(layout, expand=True, fill=True)
        parent.layout.add(buttons, fill=True)
        return buttons, layout

    def add_spacer_if_needed(self):
        if self.added_spacer:
            return
        # self.layout.add_spacer(0, expand=True)
        self.added_spacer = True

    def add_button(self, button):
        self.add_spacer_if_needed()
        self.button_layout.add(button, margin_left=10, fill=True)
        return button

    def create_close_button(self):
        self.add_spacer_if_needed()
        button = Button(self, gettext("Close"))
        button.activated.connect(self.__close_activated)
        self.layout.add(button, margin_left=10, fill=True)
        return button

    def __close_activated(self):
        self.get_parent().close()



import fsui
from fsbc.application import app
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.settings.settings_page import SettingsPage


class AdvancedSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("settings", "pkg:workspace")
        title = gettext("Advanced")
        subtitle = gettext(
            "Specify global options and settings which does "
            "not have UI controls"
        )
        self.add_header(icon, title, subtitle)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "You can write key = value pairs here to set FS-UAE options "
                "not currently supported by the user interface. This is only a "
                "temporary feature until the GUI supports all options "
                "directly."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_bottom=10)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "The options specified here are global and will apply to all "
                "configurations. Config options such as hardware and memory "
                "options will be ignored. Options suitable here are options "
                "like theme options."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_bottom=10)

        self.text_area = fsui.TextArea(self, font_family="monospace")
        self.text_area.set_text(self.get_initial_text())
        self.layout.add(self.text_area, fill=True, expand=True)
        self.text_area.changed.connect(self.update_settings)

    def update_settings(self):
        text = self.text_area.get_text()
        # FIXME: accessing values directly here, not very nice
        keys = list(app.settings.values.keys())
        for key in keys:
            if key not in LauncherSettings.default_settings:
                LauncherSettings.set(key, "")
                del app.settings.values[key]

        for line in text.split("\n"):
            line = line.strip()
            parts = line.split("=", 1)
            if len(parts) == 2:
                key = parts[0].strip()
                # if key in Settings.default_settings:
                #     continue
                value = parts[1].strip()
                app.settings[key] = value

    @staticmethod
    def get_initial_text():
        text = ""
        # FIXME: accessing values directly here, not very nice
        keys = app.settings.values.keys()
        for key in sorted(keys):
            if key in LauncherSettings.default_settings:
                continue
            # if key in LauncherConfig.config_keys:
            #     # print("(settings) ignoring key", key)
            #     text += "\n# {0} is ignored here " \
            #             "(use config dialog instead)\n".format(key)
            if LauncherConfig.is_config_only_option(key):
                text += (
                    "\n# {0} will here function as a global config "
                    "default and may cause\nunexpected problems. It is "
                    "recommended to only use this as a per-config "
                    "option.\n".format(key)
                )
            value = app.settings[key]
            if LauncherConfig.get(key):
                text += (
                    "\n# {0} is overridden by current "
                    "configuration\n".format(key)
                )
            text += "{0} = {1}\n".format(key, value)
            if LauncherConfig.get(key):
                text += "\n"
            if key in LauncherConfig.config_keys:
                text += "\n"
        return text



import re
import functools
from typing import List, Union, Any


def unused(_: Any) -> None:
    pass


def is_uuid(value: str) -> str:
    """
    >>> is_uuid("2c244488-4a91-4c21-a0b7-0f731d4336c7")
    '2c244488-4a91-4c21-a0b7-0f731d4336c7'
    >>> is_uuid("test")
    ''
    """
    match = re.match(
        "^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        value.lower(),
    )
    if match is not None:
        return match.group(0)
    return ""


def memoize(func):
    memoize_dict = {}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            key = (args, frozenset(kwargs.items()))
        except TypeError:
            # cannot create key -- for instance, passing a list as an argument.
            # FIXME: Log warning here
            return func(*args, **kwargs)
        try:
            return memoize_dict[key]
        except KeyError:
            value = func(*args, **kwargs)
            try:
                memoize_dict[key] = value
            except TypeError:
                # not cacheable -- for instance, passing a list as an
                # argument.
                # FIXME: Log warning here
                # FIXME: will not happen here.. se above type error?
                pass
            return value

    return wrapper


def split_version(version_string: str) -> List[str]:
    pattern = re.compile(
        "^([0-9]{1,4})(?:\.([0-9]{1,4}))?"
        "(?:\.([0-9]{1,4}))?(?:\.([0-9]{1,4}))?"
        "(~?[a-z0-9\.]*)?(?:[-_]([0-9]+))?$"
    )
    m = pattern.match(version_string)
    if m is None:
        raise ValueError(version_string + " is not a valid version number")
    return list(m.groups())


class Version(object):
    def __init__(self, version_string: str) -> None:
        self.string = version_string
        v = split_version(version_string)
        self.major = int(v[0])
        self.minor = int(v[1] or "0")
        self.revision = int(v[2] or "0")
        self.build = int(v[3] or "0")
        self.val = self.major * 10000 ** 3
        self.val += self.minor * 10000 ** 2
        self.val += self.revision * 10000 ** 1
        self.val += self.build * 10000 ** 0
        self.mod = v[4]
        self.release = -1 if v[5] is None else int(v[5])
        # print(
        #     {
        #         "major": self.major,
        #         "minor": self.minor,
        #         "revision": self.revision,
        #         "build": self.build,
        #         "val": self.val,
        #         "mod": self.mod,
        #         "release": self.release,
        #     }
        # )

    def cmp_value(self):
        mod_cmp = self.mod or "~~e"
        if not mod_cmp.startswith("~"):
            mod_cmp = "~~" + mod_cmp
        return self.val, mod_cmp, self.release

    def __lt__(self, other: "Version") -> bool:
        return self.cmp_value() < other.cmp_value()

    def __gt__(self, other: "Version") -> bool:
        return self.cmp_value() > other.cmp_value()

    def __str__(self) -> str:
        return self.string


def compare_versions(a: Union[Version, str], b: Union[Version, str]):
    """
    >>> compare_versions("1.0.0", "1.0.1")
    -1
    >>> compare_versions("2.0.0", "1.0.1")
    1
    >>> compare_versions("1.2.3u2", "1.2.3")
    1
    >>> compare_versions("1.2.3a2", "1.2.3")
    -1
    >>> compare_versions("2.0.0", "2.0.0beta3")
    1
    >>> compare_versions("2.0.1", "2.0.0~beta3")
    1
    >>> compare_versions("2.0.0", "2.0.0~beta3")
    1
    >>> compare_versions("2.5.30~dev", "2.5.30")
    -1
    >>> compare_versions("2.5.30~dev2", "2.5.30~dev1")
    1
    >>> compare_versions("2.6beta", "2.5.30~dev1")
    1
    >>> compare_versions("2.6beta", "2.6")
    -1
    >>> compare_versions("2.6beta", "2.6.0")
    -1
    >>> compare_versions("2.6beta", "2.6.1")
    -1
    >>> compare_versions("2.6.0beta1", "2.6.1")
    -1
    >>> compare_versions("5.9.1.1", "5.9.1dev")
    1
    >>> compare_versions("5.9.1.1", "5.9.2dev")
    -1
    >>> compare_versions("3.8.1qemu2.2.0", "3.9.1qemu2.2.0")
    -1
    >>> compare_versions("3.8.1qemu2.4.0", "3.8.1qemu2.2.0")
    1
    >>> compare_versions("1.22.2-1", "1.22.2")
    1
    >>> compare_versions("1.22.2-0", "1.22.2")
    1
    >>> compare_versions("3.3-0", "3.4")
    -1
    >>> compare_versions("3.3-0", "3.3-1")
    -1
    >>> compare_versions("3.3~fs0", "3.4")
    -1
    >>> compare_versions("3.3~fs0", "3.3~fs1")
    -1
    >>> compare_versions("1.22.2fs1", "1.22.2")
    1
    >>> compare_versions("1.22.2fs0", "1.22.2")
    1
    """
    # >>> compare_versions("2.6.0beta1", "2.6.0beta1.1")
    # -1
    # >>> compare_versions("2.6.0beta2", "2.6.0beta1.1")
    # 1
    if isinstance(a, Version):
        pass
    elif isinstance(a, str):
        a = Version(a)
    else:
        raise TypeError("Not a valid version string or object")
    if isinstance(b, Version):
        pass
    elif isinstance(b, str):
        b = Version(b)
    else:
        raise TypeError("Not a valid version string or object")
    # cmp is gone in Python 3
    return (a > b) - (a < b)


def chained_exception(new_e, org_e):
    new_e.__cause__ = org_e
    return new_e


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from fsui.common.element import Element, LightElement
from fsui.common.group import Group
from fsui.common.layout import Layout, VerticalLayout, HorizontalLayout
from fsui.common.spacer import Spacer
from fsui.qt import *
from fsui.qt.adapter import Adapter
from fsui.qt.Application import Application
from fsui.qt.Button import Button
from fsui.qt.CheckBox import CheckBox, HeadingCheckBox
from fsui.qt.Color import Color
from fsui.qt.Choice import Choice, ItemChoice
from fsui.qt.ComboBox import ComboBox
from fsui.qt.dialog import Dialog
from fsui.qt.dialogbuttons import DialogButtons
from fsui.qt.dialogwindow import DialogWindow
from fsui.qt.DirDialog import DirDialog
from fsui.qt.DrawingContext import DrawingContext, Font
from fsui.qt.FileDialog import (
    FileDialog,
    pick_directory,
    pick_file,
    pick_files,
)
from fsui.qt.HeadingLabel import HeadingLabel
from fsui.qt.Icon import Icon
from fsui.qt.Image import Image
from fsui.qt.ImageButton import ImageButton
from fsui.qt.ImageView import ImageView
from fsui.qt.label import Label, PlainLabel, URLLabel, MultiLineLabel
from fsui.qt.legacydialog import LegacyDialog
from fsui.qt.ListView import ListView
from fsui.qt.Menu import Menu, PopupMenu
from fsui.qt.Panel import Panel
from fsui.qt.scrollarea import VerticalScrollArea
from fsui.qt.SpinCtrl import SpinCtrl
from fsui.qt.System import System
from fsui.qt.timer import IntervalTimer
from fsui.qt.TextArea import TextArea
from fsui.qt.TextField import TextField, PasswordField
from fsui.qt.VerticalItemView import VerticalItemView
from fsui.qt.window import Window


default_window_parent = []
default_window_center = []
toolkit = "qt"
use_qt = True
theme = ""
theme_variant = ""


# noinspection PyProtectedMember
def open_window_instance(cls, parent=None):
    if not hasattr(cls, "_window_instance"):
        cls._window_instance = None
    if cls._window_instance is not None:
        cls._window_instance.raise_and_activate()
        return cls._window_instance
    cls._window_instance = cls(parent)

    def reset_instance():
        print("SettingsDialog.reset_instance")
        # cls._window_instance.deleteLater()
        cls._window_instance = None

    cls._window_instance.closed.connect(reset_instance)
    cls._window_instance.show()
    return cls._window_instance

    # def monitor_instance_2(count):
    #     if count < 100:
    #         call_after(monitor_instance_2, count + 1)
    #         return
    #     print("DESTROYED SIGNAL RECEIVED 2")
    #     instance = weak_instance()
    #     if instance is not None:
    #         print("WARNING: SettingsDialog is still alive")
    #         import gc
    #         print(gc.get_referrers(instance))
    #         print("real window:")
    #         print(gc.get_referrers(instance.real_window()))
    #     else:
    #         print("Instance is now", instance)
    #
    # def monitor_instance():
    #     print("DESTROYED SIGNAL RECEIVED")
    #     call_after(monitor_instance_2, 1)
    #
    # weak_instance = weakref.ref(cls._window_instance)
    # # cls._window_instance.destroyed.connect(monitor_instance)
    # cls._window_instance.closed.connect(monitor_instance)

    # if getattr(cls, "_window_instance", None) and cls._window_instance():
    #     cls._window_instance().raise_and_activate()
    # else:
    #     window = cls(parent)
    #     window.show()
    #     cls._window_instance = weakref.ref(window)


# noinspection PyProtectedMember
def current_window_instance(cls):
    if not hasattr(cls, "_window_instance"):
        return None
    return cls._window_instance



import fsui.qt


class IntervalTimer(fsui.qt.QObject):

    activated = fsui.qt.QSignal()

    def __init__(self, interval):
        super().__init__()
        self.startTimer(interval)

    def __del__(self):
        print("IntervalTimer.__del__", self)

    # noinspection PyPep8Naming
    def timerEvent(self, _):
        self.activated.emit()

    def stop(self):
        print("[TIMER] Stop")
        self.activated.disconnect()



import weakref
from fsui.qt import Qt, QScrollArea, QFrame
from .widget_mixin import WidgetMixin


class VerticalScrollArea(QScrollArea, WidgetMixin):
    def __init__(self, parent):
        QScrollArea.__init__(self, parent.get_container())
        self.init_widget(parent)
        self.__child = None
        # self.layout = None
        # self._painter = None

        self.move(0, 2000)
        # self.setAutoFillBackground(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    @property
    def size(self):
        return self.get_size()

    def showEvent(self, event):
        self.on_resize()

    def resizeEvent(self, event):
        QScrollArea.resizeEvent(self, event)
        self.on_resize()

    def get_child(self):
        # child = self.__child()
        # return child
        return self.__child

    def get_min_width(self):
        return self.get_child().get_min_width()

    def on_resize(self):
        child = self.get_child()
        height = child.get_min_height()
        child.set_size((self.viewport().width(), height))

    def set_widget(self, child):
        # self.__child = weakref.ref(child)
        self.__child = child
        widget = child.widget()
        self.setWidget(widget)
        widget.move(0, 0)
        self.on_resize()



from typing import Tuple
import warnings

from fsui.common.layout import Layout
from fsui.qt import QPoint
from fsui.qt import Qt, QObject, QFontMetrics, QWidget, QPalette, QCursor
from fsui.qt.Color import Color
from fsui.qt.DrawingContext import Font
from fsui.qt.signal import Signal


# noinspection PyPep8Naming
class Widget(QObject):

    destroyed = Signal()

    def __init__(self, parent, *_):
        super().__init__()
        self._parent = None
        # noinspection PyProtectedMember
        self._window = parent._window
        self._widget = None
        self._explicitly_hidden = False

    def widget(self):
        return self._widget

    def get_container(self):
        return self.widget()

    def set_widget(self, widget):
        self._widget = widget
        # self.init_mixin_base()
        # self.layout = None
        # self._parent = weakref.ref(parent)
        # noinspection PyProtectedMember
        # self._window = parent._window
        # widget = getattr(self, "_widget", self)
        # widget.move(10000, 10000)

        # if self.get_window() != widget:
        # assert isinstance(widget, QWidget)

        widget.installEventFilter(self)
        widget.destroyed.connect(self.on_destroy)

    def on_destroy(self):
        print("Widget.on_destroy", self)
        self.destroyed.emit()

    def explicitly_hidden(self):
        return self._explicitly_hidden

    def set_visible(self, show=True):
        if show:
            self.widget().show()
        else:
            self.widget().hide()
        self._explicitly_hidden = not show

    def show(self):
        self.set_visible(True)

    def hide(self):
        self.set_visible(False)

    def eventFilter(self, obj, event):
        return False

    def parent(self):
        if self._parent is None:
            return None
        # noinspection PyCallingNonCallable
        return self._parent()

    def font(self):
        return Font(self.widget().font())

    def set_font(self, font):
        self.widget().setFont(font.font)

    def is_visible(self):
        return self.widget().isVisible()

    def measure_text(self, text):
        font = self.widget().font()
        metrics = QFontMetrics(font)
        return metrics.width(text), metrics.height()

    def set_hand_cursor(self):
        self.widget().setCursor(Qt.PointingHandCursor)

    def set_normal_cursor(self):
        self.widget().setCursor(Qt.ArrowCursor)

    def is_enabled(self):
        return self.widget().isEnabled()

    def focus(self):
        self.widget().setFocus()

    def refresh(self):
        return self.widget().update()

    def set_min_size(self, size):
        # noinspection PyAttributeOutsideInit
        self.min_width = size[0]
        # noinspection PyAttributeOutsideInit
        self.min_height = size[1]

    def set_min_width(self, width):
        # noinspection PyAttributeOutsideInit
        self.min_width = width

    def set_min_height(self, height):
        # noinspection PyAttributeOutsideInit
        self.min_height = height

    def get_min_width(self):
        widget = getattr(self, "_widget", self)
        width = 0
        if hasattr(self, "min_width"):
            if self.min_width:
                width = max(self.min_width, width)
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            width = max(self.layout.get_min_width(), width)
            return width
        # result = max(width, widget.minimumSizeHint().width())
        # if widget.maximumWidth():
        #     print(widget.maximumWidth())
        #     return min(result, widget.maximumWidth())
        # return min(result, widget.maximumWidth())
        # return result
        result = max(width, widget.minimumSizeHint().width())
        return min(result, widget.maximumWidth())
        # return max(width, widget.minimumWidth())

    def get_min_height(self):
        widget = getattr(self, "_widget", self)
        assert isinstance(widget, QWidget)
        height = 0
        if hasattr(self, "min_height"):
            if self.min_height:
                height = max(self.min_height, height)
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            height = max(self.layout.get_min_height(), height)
            return height
        return max(height, widget.minimumSizeHint().height())
        # return max(height, widget.minimumHeight())

    def set_position(self, position, y=None):
        if y is None:
            self.widget().move(*position)
        else:
            self.widget().move(position, y)

    def set_size(self, size):
        self.widget().resize(*size)

    def set_position_and_size(self, position, size):
        self.widget().setGeometry(position[0], position[1], size[0], size[1])

    @property
    def window(self):
        # noinspection PyCallingNonCallable
        return self._window()

    def get_window(self):
        warnings.warn("use window property instead", DeprecationWarning)
        # noinspection PyCallingNonCallable
        return self._window()

    def position(self):
        return self.widget().x(), self.widget().y()

    def size(self):
        return self.widget().width(), self.widget().height()

    def width(self):
        return self.size()[0]

    def height(self):
        return self.size()[1]

    def set_tool_tip(self, tool_tip):
        widget = getattr(self, "_widget", self)
        widget.setToolTip(tool_tip)

    def set_tooltip(self, tool_tip):
        self.set_tool_tip(tool_tip)

    def disable(self):
        return self.enable(False)

    def enable(self, enable=True):
        widget = getattr(self, "_widget", self)
        widget.setEnabled(enable)

    def set_enabled(self, enable=True):
        self.enable(enable)

    def on_resize(self):
        if hasattr(self, "layout") and isinstance(self.layout, Layout):
            self.layout.set_size(self.get_size())
            self.layout.update()

    def get_background_color(self):
        # noinspection PyUnresolvedReferences
        return Color(self.widget().palette().color(QPalette.Window))

    def set_background_color(self, color):
        widget = self.widget()
        widget.setAutoFillBackground(True)
        p = widget.palette()
        p.setColor(widget.backgroundRole(), color)
        widget.setPalette(p)

    def popup_menu(self, menu, pos=(0, 0), blocking=True):
        # popup does not block, and if menu goes out of the scope of the
        # caller, it will disappear (unless we keep a reference here
        # FIXME: using exec now
        # self.__last_popup_menu = menu
        widget = getattr(self, "_widget", self)
        global_pos = widget.mapToGlobal(QPoint(pos[0], pos[1]))
        menu.set_parent(self)
        if blocking:
            menu.qmenu.exec(global_pos)
        else:
            menu.qmenu.popup(global_pos)

        # Firing off a fake mouse left up event really assumes that the
        # menu was opened with a left down event, so implementation isn't
        # ideal. Better to add a listener to menu close instead.

        # if hasattr(self, "on_left_up"):
        #     self.on_left_up()

    def is_mouse_over(self):
        # # noinspection PyArgumentList
        # c = QCursor.pos()
        # # noinspection PyUnresolvedReferences
        # p = self.widget().mapToGlobal(QPoint(0, 0))
        # s = self.size()
        # print(c, p, s)
        # if p.x() <= c.x() < p.x() + s[0] and p.y() <= c.y() < p.y() + s[1]:
        #     return True
        # return False
        return self.widget().underMouse()

    # DEPRECATED
    def get_parent(self):
        return self.parent()

    # DEPRECATED
    def visible(self):
        return self.is_visible()

    # DEPRECATED
    def get_font(self):
        return self.font()

    # DEPRECATED
    def show_or_hide(self, show=True):
        self.set_visible(show)

    # DEPRECATED
    def get_position(self):
        return self.position()

    # DEPRECATED
    def get_size(self) -> Tuple[int, int]:
        return self.size()



# Automatically generated - do not edit by hand


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _accelerator(c, f):
    if c.accelerator.explicit:
        value = c.accelerator.explicit
    elif c.amiga_model == "A1200/1230":
        value = "blizzard-1230-iv"
    elif c.amiga_model == "A1200/1240":
        value = "blizzard-1240"
    elif c.amiga_model == "A1200/1260":
        value = "blizzard-1260"
    elif c.amiga_model == "A1200/PPC":
        value = "blizzard-ppc"
    elif c.amiga_model == "A4000/PPC":
        value = "cyberstorm-ppc"
    elif c.amiga_model == "A4000/OS4":
        value = "cyberstorm-ppc"
    else:
        value = "0"
    if f.matches(value, "blizzard-1230-iv"):
        value = "blizzard-1230-iv"
    # elif f.matches(value, "blizzard-1230-iv-scsi"):
    #     value = "blizzard-1230-iv-scsi"
    # elif f.matches(value, "blizzard-1230-iv+scsi"):
    #     value = "blizzard-1230-iv-scsi"
    elif f.matches(value, "blizzard-1240"):
        value = "blizzard-1240"
    elif f.matches(value, "blizzard-1260"):
        value = "blizzard-1260"
    # elif f.matches(value, "blizzard-1260-scsi"):
    #     value = "blizzard-1260-scsi"
    # elif f.matches(value, "blizzard-2060"):
    #     value = "blizzard-2060"
    elif f.matches(value, "blizzard-ppc"):
        value = "blizzard-ppc"
    # elif f.matches(value, "cyberstorm-mk-i"):
    #     value = "cyberstorm-mk-i"
    # elif f.matches(value, "cyberstorm-mk-ii"):
    #     value = "cyberstorm-mk-ii"
    # elif f.matches(value, "cyberstorm-mk-iii"):
    #     value = "cyberstorm-mk-iii"
    elif f.matches(value, "cyberstorm-ppc"):
        value = "cyberstorm-ppc"
    # elif f.matches(value, "dkb-1230"):
    #     value = "dkb-1230"
    # elif f.matches(value, "dkb-1240"):
    #     value = "dkb-1240"
    # elif f.matches(value, "fusion-forty"):
    #     value = "fusion-forty"
    # elif f.matches(value, "warp-engine-a4000"):
    #     value = "warp-engine-a4000"
    c.accelerator = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _accelerator_memory(c, f):
    if c.accelerator_memory.explicit:
        value = c.accelerator_memory.explicit
    elif c.accelerator == "":
        value = "0"
    elif f.matches(c.accelerator, "blizzard-1230-iv"):
        value = str(32 * 1024)
    elif f.matches(c.accelerator, "blizzard-1240"):
        value = str(32 * 1024)
    elif f.matches(c.accelerator, "blizzard-1260"):
        value = str(32 * 1024)
    elif f.matches(c.accelerator, "blizzard-ppc"):
        value = str(256 * 1024)
    elif f.matches(c.accelerator, "cyberstorm-ppc"):
        value = str(128 * 1024)
    elif f.matches(c.accelerator, "0"):
        value = str(0 * 1024)
    else:
        f.warning("accelerator_memory")
        value = "0"
    c.accelerator_memory = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _amiga_model(c, f):
    if c.amiga_model.explicit:
        c.amiga_model = c.amiga_model.explicit
    else:
        c.amiga_model = "A500"


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _bsdsocket_library(c, f):
    if c.bsdsocket_library.explicit:
        value = c.bsdsocket_library.explicit
    else:
        value = "0"
    c.bsdsocket_library = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _cdrom_drive_0(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _cdrom_drive_count(c, f):
    if c.cdrom_drive_count.explicit:
        value = c.cdrom_drive_count.explicit
    elif c.cdrom_drive_0:
        value = "1"
    elif c.cdrom_image_0:
        value = "1"
    elif c.int_model == "CD32":
        value = "1"
    elif c.int_model == "CDTV":
        value = "1"
    else:
        value = "0"
    c.cdrom_drive_count = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _cdrom_image_0(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _chip_memory(c, f):
    if c.chip_memory.explicit:
        value = c.chip_memory.explicit
    elif c.amiga_model == "A500":
        value = "512"
    elif c.amiga_model == "A500/512K":
        value = "512"
    elif c.amiga_model == "A500+":
        value = "1024"
    elif c.amiga_model == "A600":
        value = "1024"
    elif c.amiga_model == "A1000":
        value = "512"
    elif c.amiga_model == "A1200":
        value = "2048"
    elif c.amiga_model == "A1200/3.0":
        value = "2048"
    elif c.amiga_model == "A1200/020":
        value = "2048"
    elif c.amiga_model == "A1200/1230":
        value = "2048"
    elif c.amiga_model == "A1200/1240":
        value = "2048"
    elif c.amiga_model == "A1200/1260":
        value = "2048"
    elif c.amiga_model == "A1200/PPC":
        value = "2048"
    elif c.amiga_model == "A3000":
        # FIXME: did it usually have 1 MB?
        value = "2048"
    elif c.amiga_model == "A4000":
        value = "2048"
    elif c.amiga_model == "A4000/040":
        value = "2048"
    elif c.amiga_model == "A4000/OS4":
        value = "2048"
    elif c.amiga_model == "A4000/PPC":
        value = "2048"
    elif c.amiga_model == "CD32":
        value = "2048"
    elif c.amiga_model == "CD32/FMV":
        value = "2048"
    elif c.amiga_model == "CDTV":
        value = "512"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")
    c.chip_memory = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _cpu(c, f):
    if c.cpu.explicit:
        c.cpu = c.cpu.explicit
    elif c.accelerator == "blizzard-1230-iv":
        c.cpu = "68EC030"
    elif c.accelerator == "blizzard-1240":
        c.cpu = "68040-NOMMU"
    elif c.accelerator == "blizzard-1260":
        c.cpu = "68060-NOMMU"
    elif c.accelerator == "blizzard-ppc":
        c.cpu = "68060-NOMMU"
    elif c.accelerator == "cyberstorm-ppc":
        c.cpu = "68060-NOMMU"
    elif c.amiga_model == "A500":
        c.cpu = "68000"
    elif c.amiga_model == "A500/512K":
        c.cpu = "68000"
    elif c.amiga_model == "A500+":
        c.cpu = "68000"
    elif c.amiga_model == "A600":
        c.cpu = "68000"
    elif c.amiga_model == "A1000":
        c.cpu = "68000"
    elif c.amiga_model == "A1200":
        c.cpu = "68EC020"
    elif c.amiga_model == "A1200/3.0":
        c.cpu = "68EC020"
    elif c.amiga_model == "A1200/020":
        c.cpu = "68020"
    elif c.amiga_model == "A1200/1230":
        c.cpu = "68030"
    elif c.amiga_model == "A1200/1240":
        c.cpu = "68040-NOMMU"
    elif c.amiga_model == "A1200/1260":
        c.cpu = "68060-NOMMU"
    elif c.amiga_model == "A1200/PPC":
        c.cpu = "68060-NOMMU"
    elif c.amiga_model == "A3000":
        c.cpu = "68030"
    elif c.amiga_model == "A4000":
        c.cpu = "68EC030"
    elif c.amiga_model == "A4000/040":
        c.cpu = "68040-NOMMU"
    elif c.amiga_model == "A4000/OS4":
        c.cpu = "68060-NOMMU"
    elif c.amiga_model == "A4000/PPC":
        c.cpu = "68060-NOMMU"
    elif c.amiga_model == "CD32":
        c.cpu = "68EC020"
    elif c.amiga_model == "CD32/FMV":
        c.cpu = "68EC020"
    elif c.amiga_model == "CDTV":
        c.cpu = "68000"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _dongle_type(c, f):
    # FIXME: uae_dongle_type
    if c.dongle_type.explicit:
        c.dongle_type = c.dongle_type.explicit
    else:
        c.dongle_type = "0"


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _fast_memory(c, f):
    if c.fast_memory.explicit:
        value = c.fast_memory.explicit
    else:
        value = "0"
    c.fast_memory = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_0(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_1(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_2(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_3(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_count(c, f):
    if c.floppy_drive_count.explicit:
        value = c.floppy_drive_count.explicit
    elif c.floppy_drive_3:
        value = "4"
    elif c.floppy_drive_2:
        value = "3"
    elif c.floppy_drive_1:
        value = "2"
    elif c.int_model == "A4000":
        value = "2"
    elif c.floppy_drive_0:
        value = "1"
    elif c.int_model == "CD32":
        value = "0"
    elif c.int_model == "CDTV":
        value = "0"
    elif c.platform == "amiga":
        value = "1"
    elif c.platform == "cd32":
        value = "1"
    elif c.platform == "cdtv":
        value = "1"
    elif c.platform == "":
        value = "1"
    else:
        value = "0"
    c.floppy_drive_count = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _floppy_drive_speed(c, f):
    if c.floppy_drive_speed.explicit:
        c.floppy_drive_speed = c.floppy_drive_speed.explicit
    else:
        c.floppy_drive_speed = "100"


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _fpu(c, f):
    if c.fpu.explicit:
        value = c.fpu.explicit
    elif c.amiga_model == "A3000" and not c.cpu.explicit:
        value = "68882"
    elif c.amiga_model == "A4000" and not c.cpu.explicit:
        value = "68882"
    elif c.cpu == "68000":
        value = "0"
    elif c.cpu == "68010":
        value = "0"
    elif c.cpu == "68EC020":
        value = "0"
    elif c.cpu == "68020":
        value = "0"
    elif c.cpu == "68EC030":
        value = "0"
    elif c.cpu == "68030":
        value = "0"
    elif c.cpu == "68EC040":
        value = "0"
    elif c.cpu == "68LC040":
        value = "0"
    elif c.cpu == "68040-NOMMU":
        value = "68040"
    elif c.cpu == "68040":
        value = "68040"
    elif c.cpu == "68EC060":
        value = "0"
    elif c.cpu == "68LC060":
        value = "0"
    elif c.cpu == "68060-NOMMU":
        value = "68060"
    elif c.cpu == "68060":
        value = "68060"
    else:
        f.fail("Unknown CPU")
        raise Exception("Failed")
    c.fpu = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _freezer_cartridge(c, f):
    if c.freezer_cartridge.explicit:
        c.freezer_cartridge = c.freezer_cartridge.explicit
        # FIXME: Check valid values
    else:
        c.freezer_cartridge = "0"


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _graphics_card(c, f):
    if c.graphics_card.explicit:
        # FIXME: check supported
        value = c.graphics_card.explicit
    else:
        value = ""
    c.graphics_card = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _graphics_memory(c, f):
    if c.graphics_memory.explicit:
        value = c.graphics_memory.explicit
    elif c.uae_gfxcard_size == "0":
        value = "0"
    else:
        value = int(c.uae_gfxcard_size) * 1024
    c.graphics_memory = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_accelerator_name(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_accelerator_name.explicit:
        f.fail("int_accelerator_name was set explicitly")
    if f.matches(c.uae_cpuboard_type, "none"):
        value = ""
    elif f.matches(c.uae_cpuboard_type, "Blizzard1230IV"):
        value = "Blizzard 1230 IV"
    elif f.matches(c.uae_cpuboard_type, "Blizzard1260"):
        value = "Blizzard 1240/1260"
    elif f.matches(c.uae_cpuboard_type, "Blizzard2060"):
        value = "Blizzard 2060"
    elif f.matches(c.uae_cpuboard_type, "BlizzardPPC"):
        value = "Blizzard PPC"
    elif f.matches(c.uae_cpuboard_type, "CyberStormMK1"):
        value = "CyberStorm MK I"
    elif f.matches(c.uae_cpuboard_type, "CyberStormMK2"):
        value = "CyberStorm MK II"
    elif f.matches(c.uae_cpuboard_type, "CyberStormMK3"):
        value = "CyberStorm MK III"
    elif f.matches(c.uae_cpuboard_type, "CyberStormPPC"):
        value = "Cyberstorm PPC"
    elif f.matches(c.uae_cpuboard_type, "WarpEngineA4000"):
        value = "Warp Engine"
    elif f.matches(c.uae_cpuboard_type, "TekMagic"):
        value = "Tek Magic"
    else:
        value = "Unknown Accelerator (?)"
        f.warning("Unknown accelerator")
    c.int_accelerator_name = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_bogomem_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_bogomem_size.explicit:
        f.fail("int_bogomem_size was set explicitly")
    if c.slow_memory:
        value = str(int(c.slow_memory) * 1024)
    else:
        value = "0"
    if value == "0":
        pass
    elif value == "524288":
        pass
    elif value == "1048576":
        pass
    elif value == "1572864":
        pass
    elif value == "1835008":
        pass
    else:
        f.warning("Unsupported slow memory size: " + value)
    c.int_bogomem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_chipmem_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_chipmem_size.explicit:
        f.fail("int_chipmem_size was set explicitly")
    if c.chip_memory:
        value = str(int(c.chip_memory) * 1024)
    else:
        value = "0"
    if value == "0":
        pass
    elif value == "262144":
        pass
    elif value == "524288":
        pass
    elif value == "1048576":
        pass
    elif value == "1572864":
        pass
    elif value == "2097152":
        pass
    elif value == "4194304":
        pass
    elif value == "8388608":
        pass
    else:
        f.warning("Unsupported chip memory size: " + value)
    c.int_chipmem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_chipset_name(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_chipset_name.explicit:
        f.fail("int_chipset_name was set explicitly")
    t = f.lower(c.uae_chipset)
    if t == "ocs":
        value = "OCS"
    elif t == "ecs":
        value = "ECS"
    elif t == "ecs_agnus":
        value = "ECS Agnus"
    elif t == "ecs_denise":
        value = "ECS Denise"
    elif t == "aga":
        value = "AGA"
    else:
        value = "Unknown"
        f.warning("Unknown chipset")
    c.int_chipset_name = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_cpu_name(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_cpu_name.explicit:
        f.fail("int_cpu_name was set explicitly")
    if f.matches(c.uae_cpu_model, "68000"):
        value = "68000"
    elif f.matches(c.uae_cpu_model, "68010"):
        value = "68010"
    elif f.matches(c.uae_cpu_model, "68020"):
        if f.matches(c.uae_cpu_24bit_addressing, "true"):
            value = "68EC020"
        else:
            value = "68020"
    elif f.matches(c.uae_cpu_model, "68030"):
        if f.matches(c.uae_mmu_model, "68030"):
            value = "68030"
        else:
            value = "68EC030"
    elif f.matches(c.uae_cpu_model, "68040"):
        if f.matches(c.uae_fpu_model, "68040"):
            value = "68040"
        elif f.matches(c.uae_mmu_model, "68040"):
            value = "68LC040"
        else:
            value = "68EC040"
    elif f.matches(c.uae_cpu_model, "68060"):
        if f.matches(c.uae_fpu_model, "68060"):
            value = "68060"
        elif f.matches(c.uae_mmu_model, "68060"):
            value = "68LC060"
        else:
            value = "68EC060"
    else:
        f.fail("Unknown CPU")
        raise Exception("Failed")
    c.int_cpu_name = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_cpuboardmem1_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_cpuboardmem1_size.explicit:
        f.fail("int_cpuboardmem1_size was set explicitly")
    if c.accelerator_memory:
        value = str(int(c.accelerator_memory) * 1024)
    else:
        value = "0"
    c.int_cpuboardmem1_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_default_floppy_type(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_default_floppy_type.explicit:
        f.fail("int_default_floppy_type was set explicitly")
    if c.int_model == "A3000":
        value = "1"
    elif c.int_model == "A4000":
        value = "1"
    else:
        value = "0"
    c.int_default_floppy_type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_fastmem_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_fastmem_size.explicit:
        f.fail("int_fastmem_size was set explicitly")
    if c.fast_memory:
        value = str(int(c.fast_memory) * 1024)
    else:
        value = "0"
    if value == "0":
        pass
    elif value == "65536":
        pass
    elif value == "131072":
        pass
    elif value == "262144":
        pass
    elif value == "524288":
        pass
    elif value == "1048576":
        pass
    elif value == "2097152":
        pass
    elif value == "4194304":
        pass
    elif value == "8388608":
        pass
    else:
        f.warning("Unsupported Zorro II fast memory size: " + value)
    c.int_fastmem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_graphics_card_bus(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_graphics_card_bus.explicit:
        f.fail("int_graphics_card_bus was set explicitly")
    t = c.uae_gfxcard_type
    if t == "":
        value = ""
    elif t == "PicassoII":
        value = "zorro-ii"
    elif t == "PicassoII+":
        value = "zorro-ii"
    elif t == "PicassoIV_Z2":
        value = "zorro-ii"
    elif t == "PicassoIV_Z3":
        value = "zorro-iii"
    elif t == "ZorroII":
        value = "zorro-ii"
    elif t == "ZorroIII":
        value = "zorro-iii"
    else:
        f.warning("Unrecognized graphics card: " + t)
        value = ""
    if value == "zorro-iii" and c.uae_cpu_24bit_addressing == "true":
        f.warning("Zorro III graphics card cannot be used with 24-bit CPU")
    c.int_graphics_card_bus = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_graphics_card_name(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_graphics_card_name.explicit:
        f.fail("int_graphics_card_name was set explicitly")
    t = c.uae_gfxcard_type
    if t == "PicassoII":
        value = "Picasso II"
    elif t == "PicassoII+":
        value = "Picasso II+"
    elif t == "PicassoIV_Z2":
        value = "Picasso IV"
    elif t == "PicassoIV_Z3":
        value = "Picasso IV"
    elif t == "ZorroII":
        value = "UAEGFX"
    elif t == "ZorroIII":
        value = "UAEGFX"
    else:
        value = "Unknown GFX Card"
    c.int_graphics_card_name = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_kickstart_ext_sha1(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_kickstart_ext_sha1.explicit:
        f.fail("int_kickstart_ext_sha1 was set explicitly")
    t = c.amiga_model
    if t == "A500":
        value = ""
    elif t == "A500/512K":
        value = ""
    elif t == "A500+":
        value = ""
    elif t == "A600":
        value = ""
    elif t == "A1000":
        value = ""
    elif t == "A1200":
        value = ""
    elif t == "A1200/3.0":
        value = ""
    elif t == "A1200/020":
        value = ""
    elif t == "A1200/1230":
        value = ""
    elif t == "A1200/1240":
        value = ""
    elif t == "A1200/1260":
        value = ""
    elif t == "A1200/PPC":
        value = ""
    elif t == "A3000":
        value = ""
    elif t == "A4000":
        value = ""
    elif t == "A4000/040":
        value = ""
    elif t == "A4000/OS4":
        value = ""
    elif t == "A4000/PPC":
        value = ""
    elif t == "CD32":
        value = "5bef3d628ce59cc02a66e6e4ae0da48f60e78f7f"
    elif t == "CD32/FMV":
        value = "5bef3d628ce59cc02a66e6e4ae0da48f60e78f7f"
    elif t == "CDTV":
        value = "7ba40ffa17e500ed9fed041f3424bd81d9c907be"
    else:
        f.fail("Unknown amiga_model: " + t)
        raise Exception("Failed")
    c.int_kickstart_ext_sha1 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_kickstart_revision(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_kickstart_revision.explicit:
        f.fail("int_kickstart_revision was set explicitly")
    t = c.int_kickstart_sha1
    if t == "02843c4253bbd29aba535b0aa3bd9a85034ecde4":
        # Kickstart v2.05 rev 37.350 (1992)(Commodore)(A600HD)[!]
        value = "37.350"
    elif t == "11f9e62cf299f72184835b7b2a70a16333fc0d88":
        # Kickstart v1.2 rev 33.180 (1986)(Commodore)(A500-A2000)[!]
        value = "33.180"
    elif t == "3525be8887f79b5929e017b42380a79edfee542d":
        # Kickstart v3.1 rev 40.60 (1993)(Commodore)(CD32)
        value = "40.60"
    elif t == "5fe04842d04a489720f0f4bb0e46948199406f49":
        # Kickstart v3.1 rev 40.68 (1993)(Commodore)(A4000)
        value = "40.68"
    elif t == "70033828182fffc7ed106e5373a8b89dda76faa5":
        # Kickstart v3.0 rev 39.106 (1992)(Commodore)(A1200)[!]
        value = "39.106"
    elif t == "891e9a547772fe0c6c19b610baf8bc4ea7fcb785":
        # Kickstart v1.3 rev 34.5 (1987)(Commodore)(A500-A1000-A2000-CDTV)[!]
        value = "34.5"
    elif t == "c5839f5cb98a7a8947065c3ed2f14f5f42e334a1":
        # Kickstart v2.04 rev 37.175 (1991)(Commodore)(A500+)[!]
        value = "37.175"
    elif t == "e21545723fe8374e91342617604f1b3d703094f1":
        # Kickstart v3.1 rev 40.68 (1993)(Commodore)(A1200)[!]
        value = "40.68"
    elif t == "f8e210d72b4c4853e0c9b85d223ba20e3d1b36ee":
        # Kickstart v3.1 r40.68 (1993)(Commodore)(A3000).rom
        value = "40.68"
    else:
        value = "???"
    c.int_kickstart_revision = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_kickstart_sha1(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_kickstart_sha1.explicit:
        f.fail("int_kickstart_sha1 was set explicitly")
    if c.amiga_model == "A500":
        value = "891e9a547772fe0c6c19b610baf8bc4ea7fcb785"
    elif c.amiga_model == "A500/512K":
        value = "891e9a547772fe0c6c19b610baf8bc4ea7fcb785"
    elif c.amiga_model == "A500+":
        value = "c5839f5cb98a7a8947065c3ed2f14f5f42e334a1"
    elif c.amiga_model == "A600":
        value = "02843c4253bbd29aba535b0aa3bd9a85034ecde4"
    elif c.amiga_model == "A1000":
        value = "11f9e62cf299f72184835b7b2a70a16333fc0d88"
    elif c.amiga_model == "A1200":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A1200/3.0":
        value = "70033828182fffc7ed106e5373a8b89dda76faa5"
    elif c.amiga_model == "A1200/020":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A1200/1230":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A1200/1240":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A1200/1260":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A1200/PPC":
        value = "e21545723fe8374e91342617604f1b3d703094f1"
    elif c.amiga_model == "A3000":
        value = "f8e210d72b4c4853e0c9b85d223ba20e3d1b36ee"
    elif c.amiga_model == "A4000":
        value = "5fe04842d04a489720f0f4bb0e46948199406f49"
    elif c.amiga_model == "A4000/040":
        value = "5fe04842d04a489720f0f4bb0e46948199406f49"
    elif c.amiga_model == "A4000/OS4":
        value = "5fe04842d04a489720f0f4bb0e46948199406f49"
    elif c.amiga_model == "A4000/PPC":
        value = "5fe04842d04a489720f0f4bb0e46948199406f49"
    elif c.amiga_model == "CD32":
        value = "3525be8887f79b5929e017b42380a79edfee542d"
    elif c.amiga_model == "CD32/FMV":
        value = "3525be8887f79b5929e017b42380a79edfee542d"
    elif c.amiga_model == "CDTV":
        value = "891e9a547772fe0c6c19b610baf8bc4ea7fcb785"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")
    c.int_kickstart_sha1 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_kickstart_version(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_kickstart_version.explicit:
        f.fail("int_kickstart_version was set explicitly")
    t = c.int_kickstart_revision
    if t == "33.180":
        value = "1.2"
    elif t == "34.5":
        value = "1.3"
    elif t == "37.175":
        value = "2.04"
    elif t == "37.350":
        value = "2.05"
    elif t == "39.106":
        value = "3.0"
    elif t == "40.60":
        value = "3.1"
    elif t == "40.68":
        value = "3.1"
    else:
        value = "???"
    c.int_kickstart_version = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_mbresmem_low_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_mbresmem_low_size.explicit:
        f.fail("int_mbresmem_low_size was set explicitly")
    if int(c.uae_a3000mem_size):
        value = str(int(c.uae_a3000mem_size) * 1024 * 1024)
    else:
        value = "0"
    c.int_mbresmem_low_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_model(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_model.explicit:
        f.fail("int_model was set explicitly")
    if c.amiga_model == "A500":
        c.int_model = "A500"
    elif c.amiga_model == "A500/512K":
        c.int_model = "A500"
    elif c.amiga_model == "A500+":
        c.int_model = "A500+"
    elif c.amiga_model == "A600":
        c.int_model = "A600"
    elif c.amiga_model == "A1000":
        c.int_model = "A1000"
    elif c.amiga_model == "A1200":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/3.0":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/020":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/1230":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/1240":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/1260":
        c.int_model = "A1200"
    elif c.amiga_model == "A1200/PPC":
        c.int_model = "A1200"
    elif c.amiga_model == "A3000":
        c.int_model = "A3000"
    elif c.amiga_model == "A4000":
        c.int_model = "A4000"
    elif c.amiga_model == "A4000/040":
        c.int_model = "A4000"
    elif c.amiga_model == "A4000/OS4":
        c.int_model = "A4000"
    elif c.amiga_model == "A4000/PPC":
        c.int_model = "A4000"
    elif c.amiga_model == "CD32":
        c.int_model = "CD32"
    elif c.amiga_model == "CD32/FMV":
        c.int_model = "CD32"
    elif c.amiga_model == "CDTV":
        c.int_model = "CDTV"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_model_name(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_model_name.explicit:
        f.fail("int_model_name was set explicitly")
    if c.int_model == "A500":
        value = "Amiga 500"
    elif c.int_model == "A500+":
        value = "Amiga 500+"
    elif c.int_model == "A600":
        value = "Amiga 600"
    elif c.int_model == "A1000":
        value = "Amiga 1000"
    elif c.int_model == "A1200":
        value = "Amiga 1200"
    elif c.int_model == "A3000":
        value = "Amiga 3000"
    elif c.int_model == "A4000":
        value = "Amiga 4000"
    elif c.int_model == "CD32":
        value = "Amiga CD32"
    elif c.int_model == "CDTV":
        value = "Commodore CDTV"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")
    c.int_model_name = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_ppc_model(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_ppc_model.explicit:
        f.fail("int_ppc_model was set explicitly")
    if f.matches(c.uae_cpuboard_type, "BlizzardPPC"):
        value = "603ev"
    elif f.matches(c.uae_cpuboard_type, "CyberStormPPC"):
        value = "604e"
    else:
        value = ""
    c.int_ppc_model = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_uae_boot_rom(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_uae_boot_rom.explicit:
        f.fail("int_uae_boot_rom was set explicitly")
    if c.uae_bsdsocket_emu == "true":
        value = "true"
    elif c.uae_sana2 == "true":
        value = "true"
    elif c.uae_gfxcard_type == "ZorroII" and c.uae_gfxcard_size != "0":
        value = "true"
    elif c.uae_gfxcard_type == "ZorroIII" and c.uae_gfxcard_size != "0":
        value = "true"
    elif c.uae_z3chipmem_size != "0":
        value = "true"
    else:
        value = "false"
    # FIXME: chipmem size > 2 MB
    # FIXME:
    # if (nr_directory_units (NULL))
    #     return b;
    # if (nr_directory_units (&currprefs))
    #     return b;
    # if (currprefs.uaeserial)
    #     return b;
    # if (currprefs.scsi == 1)
    #     return b;
    # if (currprefs.input_tablet > 0)
    #     return b;
    # if (currprefs.rtgmem_size && currprefs.rtgmem_type < GFXBOARD_HARDWARE)
    #     return b;
    # if (currprefs.win32_automount_removable)
    #     return b;
    # if (currprefs.chipmem_size > 2 * 1024 * 1024)
    #     return b;
    c.int_uae_boot_rom = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_z3chipmem_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_z3chipmem_size.explicit:
        f.fail("int_z3chipmem_size was set explicitly")
    if int(c.uae_z3chipmem_size):
        value = str(int(c.uae_z3chipmem_size) * 1024 * 1024)
    else:
        value = "0"
    c.int_z3chipmem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _int_z3fastmem_size(c, f):
    # noinspection PyUnresolvedReferences
    if c.int_z3fastmem_size.explicit:
        f.fail("int_z3fastmem_size was set explicitly")
    if int(c.uae_z3mem_size):
        value = str(int(c.uae_z3mem_size) * 1024 * 1024)
    # if c.zorro_iii_memory:
    #     value = str(int(c.zorro_iii_memory) * 1024)
    else:
        value = "0"
    c.int_z3fastmem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _jit_compiler(c, f):
    if c.jit_compiler.explicit:
        c.jit_compiler = c.jit_compiler.explicit
    else:
        c.jit_compiler = "0"


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _joystick_port_0_mode(c, f):
    if c.joystick_port_0_mode.explicit:
        value = c.joystick_port_0_mode.explicit
    elif c.amiga_model == "CD32":
        value = "cd32 gamepad"
    else:
        # FIXME: depends on actual device in joystick_port_0...
        value = "mouse"
    c.joystick_port_0_mode = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _joystick_port_1_mode(c, f):
    if c.joystick_port_1_mode.explicit:
        value = c.joystick_port_1_mode.explicit
    elif c.amiga_model == "CD32":
        value = "cd32 gamepad"
    else:
        # FIXME: depends on actual device in joystick_port_1...
        value = "joystick"
    c.joystick_port_1_mode = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _joystick_port_2_mode(c, f):
    if c.joystick_port_2_mode.explicit:
        value = c.joystick_port_2_mode.explicit
    else:
        # FIXME: depends on actual device in joystick_port_2...?
        value = "nothing"
    c.joystick_port_2_mode = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _joystick_port_3_mode(c, f):
    if c.joystick_port_3_mode.explicit:
        value = c.joystick_port_3_mode.explicit
    else:
        # FIXME: depends on actual device in joystick_port_3...?
        value = "nothing"
    c.joystick_port_3_mode = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _kickstart_ext_file(c, f):
    if c.kickstart_file.explicit:
        value = c.kickstart_file.explicit
    else:
        value = ""
    c.kickstart_ext_file = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _kickstart_file(c, f):
    if c.kickstart_file.explicit:
        value = c.kickstart_file.explicit
    else:
        value = ""
    c.kickstart_file = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _motherboard_ram(c, f):
    if c.motherboard_ram.explicit:
        value = c.motherboard_ram.explicit
    elif c.int_model == "A3000":
        value = "8192"
    elif c.int_model == "A4000":
        # FIXME: should PPC / OS4 models be excluded?
        value = "8192"
    else:
        value = "0"
    c.motherboard_ram = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _network_card(c, f):
    if c.network_card.explicit:
        # FIXME: check supported
        value = c.network_card.explicit
        if f.matches(value, "a2065"):
            value = "a2065"
        else:
            f.warning(value + ": invalid value")
            value = "0"
    else:
        value = "0"
    c.network_card = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _platform(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _slow_memory(c, f):
    if c.slow_memory.explicit:
        value = c.slow_memory.explicit
    elif c.amiga_model == "A500":
        value = "512"
    else:
        value = "0"
    c.slow_memory = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _sound_card(c, f):
    if c.sound_card.explicit:
        # FIXME: check supported
        value = c.sound_card.explicit
        if f.matches(value, "toccata"):
            value = "toccata"
        else:
            f.warning(value + ": invalid value")
            value = "0"
    else:
        value = "0"
    c.sound_card = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_a2065(c, f):
    if c.uae_a2065.explicit:
        # FIXME: ok? keep already specified value
        value = c.uae_a2065.explicit
        # FIXME: check value
    elif c.network_card == "a2065":
        value = "slirp"
    else:
        value = ""
    c.uae_a2065 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_a3000mem_size(c, f):
    if c.uae_a3000mem_size.explicit:
        value = c.uae_a3000mem_size.explicit
    elif c.motherboard_ram != "":
        value = str(int(c.motherboard_ram) // 1024)
    # elif c.int_model == "A3000":
    #     value = "8"
    # elif c.int_model == "A4000":
    #     # FIXME: should PPC / OS4 models be excluded?
    #     value = "8"
    else:
        value = "0"
    c.uae_a3000mem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_bogomem_size(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_bsdsocket_emu(c, f):
    if c.uae_bsdsocket_emu.explicit:
        value = c.uae_bsdsocket_emu.explicit
        # FIXME: normalize to "true" or "false"
    elif c.bsdsocket_library == "1":
        value = "true"
    else:
        value = "false"
    c.uae_bsdsocket_emu = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_cd32cd(c, f):
    if c.uae_cd32cd.explicit:
        # FIXME: boolean
        value = c.uae_cd32cd.explicit
    elif c.int_model == "CD32":
        value = "true"
    else:
        value = "false"
    c.uae_cd32cd = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_cd32fmv(c, f):
    if c.uae_cd32fmv.explicit:
        # FIXME: boolean
        value = c.uae_cd32fmv.explicit
    elif c.amiga_model == "CD32/FMV":
        value = "true"
    else:
        value = "false"
    c.uae_cd32fmv = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_chipmem_size(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_chipset(c, f):
    if c.uae_chipset.explicit:
        if int(c.int_chipmem_size) >= 1048576:
            value = "ecs_agnus"
        else:
            value = "ocs"
    elif c.amiga_model == "A500":
        if int(c.int_chipmem_size) >= 1048576:
            value = "ecs_agnus"
        else:
            value = "ocs"
    elif c.amiga_model == "A500/512K":
        value = "ocs"
    elif c.amiga_model == "A500+":
        value = "ecs"
    elif c.amiga_model == "A600":
        value = "ecs"
    elif c.amiga_model == "A1000":
        value = "ocs"
    elif c.amiga_model == "A1200":
        value = "aga"
    elif c.amiga_model == "A1200/3.0":
        value = "aga"
    elif c.amiga_model == "A1200/020":
        value = "aga"
    elif c.amiga_model == "A1200/1230":
        value = "aga"
    elif c.amiga_model == "A1200/1240":
        value = "aga"
    elif c.amiga_model == "A1200/1260":
        value = "aga"
    elif c.amiga_model == "A1200/PPC":
        value = "aga"
    elif c.amiga_model == "A3000":
        value = "ecs"
    elif c.amiga_model == "A4000":
        value = "aga"
    elif c.amiga_model == "A4000/040":
        value = "aga"
    elif c.amiga_model == "A4000/OS4":
        value = "aga"
    elif c.amiga_model == "A4000/PPC":
        value = "aga"
    elif c.amiga_model == "CD32":
        value = "aga"
    elif c.amiga_model == "CD32/FMV":
        value = "aga"
    elif c.amiga_model == "CDTV":
        value = "ecs_agnus"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")
    c.uae_chipset = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_chipset_compatible(c, f):
    if c.uae_chipset_compatible.explicit:
        value = c.uae_chipset_compatible.explicit
    elif c.amiga_model == "A500":
        value = "A500"
    elif c.amiga_model == "A500/512K":
        value = "A500"
    elif c.amiga_model == "A500+":
        value = "A500+"
    elif c.amiga_model == "A600":
        value = "A600"
    elif c.amiga_model == "A1000":
        value = "A1000"
    elif c.amiga_model == "A1200":
        value = "A1200"
    elif c.amiga_model == "A1200/3.0":
        value = "A1200"
    elif c.amiga_model == "A1200/020":
        value = "A1200"
    elif c.amiga_model == "A1200/1230":
        value = "A1200"
    elif c.amiga_model == "A1200/1240":
        value = "A1200"
    elif c.amiga_model == "A1200/1260":
        value = "A1200"
    elif c.amiga_model == "A1200/PPC":
        value = "A1200"
    elif c.amiga_model == "A3000":
        value = "A3000"
    elif c.amiga_model == "A4000":
        value = "A4000"
    elif c.amiga_model == "A4000/040":
        value = "A4000"
    elif c.amiga_model == "A4000/OS4":
        value = "A4000"
    elif c.amiga_model == "A4000/PPC":
        value = "A4000"
    elif c.amiga_model == "CD32":
        value = "CD32"
    elif c.amiga_model == "CD32/FMV":
        value = "CD32"
    elif c.amiga_model == "CDTV":
        value = "CDTV"
    else:
        f.fail("Unknown amiga_model")
        raise Exception("Failed")
    c.uae_chipset_compatible = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_cpu_24bit_addressing(c, f):
    if c.uae_cpu_24bit_addressing.explicit:
        c.uae_cpu_24bit_addressing = c.uae_cpu_24bit_addressing.explicit
    elif c.cpu == "68000":
        c.uae_cpu_24bit_addressing = "true"
    elif c.cpu == "68010":
        c.uae_cpu_24bit_addressing = "true"
    elif c.cpu == "68EC020":
        c.uae_cpu_24bit_addressing = "true"
    elif c.cpu == "68020":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68EC030":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68030":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68EC040":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68LC040":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68040-NOMMU":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68040":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68EC060":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68LC060":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68060-NOMMU":
        c.uae_cpu_24bit_addressing = "false"
    elif c.cpu == "68060":
        c.uae_cpu_24bit_addressing = "false"
    else:
        f.fail("Unknown CPU")
        raise Exception("Failed")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_cpu_model(c, f):
    if c.uae_cpu_model.explicit:
        value = c.uae_cpu_model.explicit
    elif c.cpu == "68000":
        value = "68000"
    elif c.cpu == "68010":
        value = "68010"
    elif c.cpu == "68EC020":
        value = "68020"
    elif c.cpu == "68020":
        value = "68020"
    elif c.cpu == "68EC030":
        value = "68030"
    elif c.cpu == "68030":
        value = "68030"
    elif c.cpu == "68EC040":
        value = "68040"
    elif c.cpu == "68LC040":
        value = "68040"
    elif c.cpu == "68040-NOMMU":
        value = "68040"
    elif c.cpu == "68040":
        value = "68040"
    elif c.cpu == "68EC060":
        value = "68060"
    elif c.cpu == "68LC060":
        value = "68060"
    elif c.cpu == "68060-NOMMU":
        value = "68060"
    elif c.cpu == "68060":
        value = "68060"
    else:
        f.fail("Unknown CPU")
        raise Exception("Failed")
    c.uae_cpu_model = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_cpuboard_type(c, f):
    if c.uae_cpuboard_type.explicit:
        value = c.uae_cpuboard_type.explicit
    elif f.matches(c.accelerator, "0"):
        value = "none"
    elif f.matches(c.accelerator, "blizzard-1230-iv"):
        value = "Blizzard1230IV"
    elif f.matches(c.accelerator, "blizzard-1240"):
        value = "Blizzard1260"
    elif f.matches(c.accelerator, "blizzard-1260"):
        value = "Blizzard1260"
    elif f.matches(c.accelerator, "blizzard-2060"):
        value = "Blizzard2060"
    elif f.matches(c.accelerator, "blizzard-ppc"):
        value = "BlizzardPPC"
    elif f.matches(c.accelerator, "cyberstorm-mk-i"):
        value = "CyberStormMK1"
    elif f.matches(c.accelerator, "cyberstorm-mk-ii"):
        value = "CyberStormMK2"
    elif f.matches(c.accelerator, "cyberstorm-mk-iii"):
        value = "CyberStormMK3"
    elif f.matches(c.accelerator, "cyberstorm-ppc"):
        value = "CyberStormPPC"
    elif f.matches(c.accelerator, "warp-engine-a4000"):
        value = "WarpEngineA4000"
    elif f.matches(c.accelerator, "tek-magic"):
        value = "TekMagic"
    elif f.matches(c.accelerator, "dkb-1230"):
        value = "DKB12x0"
    elif f.matches(c.accelerator, "dkb-1240"):
        value = "DKB12x0"
    elif f.matches(c.accelerator, "fusion-forty"):
        value = "FusionForty"
    else:
        f.fail("Unknown accelerator")
        raise Exception("Failed")
    c.uae_cpuboard_type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_fastmem_size(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy0(c, f):
    if c.uae_floppy0.explicit:
        value = c.uae_floppy0.explicit
    else:
        value = c.floppy_drive_0
    c.uae_floppy0 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy0type(c, f):
    if c.uae_floppy0type.explicit:
        value = c.uae_floppy0type.explicit
    elif int(c.floppy_drive_count) > 0:
        value = c.int_default_floppy_type
    else:
        value = "-1"
    c.uae_floppy0type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy1(c, f):
    if c.uae_floppy1.explicit:
        value = c.uae_floppy1.explicit
    else:
        value = c.floppy_drive_1
    c.uae_floppy1 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy1type(c, f):
    if c.uae_floppy1type.explicit:
        value = c.uae_floppy1type.explicit
    elif int(c.floppy_drive_count) > 1:
        value = c.int_default_floppy_type
    else:
        value = "-1"
    c.uae_floppy1type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy2(c, f):
    if c.uae_floppy2.explicit:
        value = c.uae_floppy2.explicit
    else:
        value = c.floppy_drive_2
    c.uae_floppy2 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy2type(c, f):
    if c.uae_floppy2type.explicit:
        value = c.uae_floppy2type.explicit
    elif int(c.floppy_drive_count) > 2:
        value = c.int_default_floppy_type
    else:
        value = "-1"
    c.uae_floppy2type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy3(c, f):
    if c.uae_floppy3.explicit:
        value = c.uae_floppy3.explicit
    else:
        value = c.floppy_drive_3
    c.uae_floppy3 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_floppy3type(c, f):
    if c.uae_floppy3type.explicit:
        value = c.uae_floppy3type.explicit
    elif int(c.floppy_drive_count) > 3:
        value = c.int_default_floppy_type
    else:
        value = "-1"
    c.uae_floppy3type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_fpu_model(c, f):
    if c.uae_fpu_model.explicit:
        value = c.uae_fpu_model.explicit
    elif c.fpu == "0":
        value = "0"
    elif c.fpu == "68881":
        value = "68881"
    elif c.fpu == "68882":
        value = "68882"
    elif c.fpu == "68040":
        value = "68040"
    elif c.fpu == "68060":
        value = "68060"
    else:
        f.fail("Unknown FPU")
        raise Exception("Failed")
    c.uae_fpu_model = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_gfxcard_size(c, f):
    t = c.uae_gfxcard_type
    if c.uae_gfxcard_size.explicit:
        value = c.uae_gfxcard_size.explicit
    elif t == "":
        value = "0"
    elif t == "PicassoII":
        value = "2"
    elif t == "PicassoII+":
        value = "2"
    elif t == "PicassoIV_Z2":
        value = "4"
    elif t == "PicassoIV_Z3":
        value = "4"
    elif t == "ZorroII":
        value = "4"
    elif t == "ZorroIII":
        value = "16"
    else:
        f.warning("Unrecognized graphics card: " + t)
        value = "0"
    c.uae_gfxcard_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_gfxcard_type(c, f):
    t = c.graphics_card
    if c.uae_gfxcard_type.explicit:
        value = c.uae_gfxcard_type.explicit
    elif t == "":
        value = ""
    elif t == "uaegfx":
        if c.uae_cpu_24bit_addressing == "true":
            value = "ZorroII"
        else:
            value = "ZorroIII"
    elif t == "uaegfx-z2":
        value = "ZorroII"
    elif t == "uaegfx-z3":
        value = "ZorroIII"
    elif t == "picasso-ii":
        value = "PicassoII"
    elif t == "picasso-ii+":
        value = "PicassoII+"
    elif t == "picasso-iv":
        if c.uae_cpu_24bit_addressing == "true":
            value = "PicassoIV_Z2"
        else:
            value = "PicassoIV_Z3"
    elif t == "picasso-iv-z2":
        value = "PicassoIV_Z2"
    elif t == "picasso-iv-z3":
        value = "PicassoIV_Z3"
    # FIXME: cards are missing here
    else:
        f.warning("Unrecognized graphics card: " + t)
        value = ""
    c.uae_gfxcard_type = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_joyport0(c, f):
    if c.uae_joyport0.explicit:
        f.warning("uae_joyport0 specified (use official options instead")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_joyport0autofire(c, f):
    if c.uae_joyport0autofire.explicit:
        f.warning("uae_joyport0autofire: use official options instead")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_joyport1(c, f):
    if c.uae_joyport1.explicit:
        f.warning("uae_joyport1 specified (use official options instead")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_joyport1autofire(c, f):
    if c.uae_joyport1autofire.explicit:
        f.warning("uae_joyport1autofire: use official options instead")


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_mmu_model(c, f):
    if c.uae_mmu_model.explicit:
        value = c.uae_mmu_model.explicit
    elif c.cpu == "68000":
        value = "0"
    elif c.cpu == "68010":
        value = "0"
    elif c.cpu == "68EC020":
        value = "0"
    elif c.cpu == "68020":
        value = "0"
    elif c.cpu == "68EC030":
        value = "0"
    elif c.cpu == "68030":
        value = "68030"
    elif c.cpu == "68EC040":
        value = "0"
    elif c.cpu == "68LC040":
        value = "68040"
    elif c.cpu == "68040-NOMMU":
        value = "0"
    elif c.cpu == "68040":
        value = "68040"
    elif c.cpu == "68EC060":
        value = "0"
    elif c.cpu == "68LC060":
        value = "68060"
    elif c.cpu == "68060-NOMMU":
        value = "0"
    elif c.cpu == "68060":
        value = "68060"
    else:
        f.fail("Unknown CPU")
        raise Exception("Failed")
    c.uae_mmu_model = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_native_code(c, f):
    if c.uae_native_code.explicit:
        value = c.uae_native_code.explicit
    elif c.uaenative_library == "1":
        value = "true"
    else:
        value = "false"
    c.uae_native_code = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_ppc_model(c, f):
    pass


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_rtc(c, f):
    # FIXME: Blizzard expansions!
    if c.uae_chipset_compatible == "-":
        if c.uae_rtc.explicit:
            value = c.uae_rtc.explicit
            if f.matches(value, "none"):
                value = "none"
            elif f.matches(value, "auto"):
                value = "none"
            elif f.matches(value, "MSM6242B"):
                value = "MSM6242B"
            elif f.matches(value, "MSM6242B_A2000"):
                value = "MSM6242B_A2000"
            elif f.matches(value, "RP5C01A"):
                value = "RP5C01A"
            else:
                f.warning(value + ": invalid value")
                value = "none"
        else:
            value = "none"
    else:
        if c.uae_rtc.explicit and c.uae_rtc.explicit != "auto":
            f.warning("uae_rtc is ignored (compatible chipset enabled)")
        if c.uae_chipset_compatible == "Generic":
            value = "RP5C01A"
        elif c.uae_chipset_compatible == "CDTV":
            value = "MSM6242B"
        elif c.uae_chipset_compatible == "CD32":
            value = "none"
        elif c.uae_chipset_compatible == "A500":
            if int(c.int_bogomem_size):
                value = "MSM6242B"
            elif int(c.int_chipmem_size) > 0x80000:
                value = "MSM6242B"
            elif int(c.int_fastmem_size):
                value = "MSM6242B"
            else:
                value = "none"
        elif c.uae_chipset_compatible == "A500+":
            value = "MSM6242B"
        elif c.uae_chipset_compatible == "A600":
            value = "none"
        elif c.uae_chipset_compatible == "A1000":
            value = "none"
        elif c.uae_chipset_compatible == "A1200":
            if int(c.int_fastmem_size) or int(c.int_z3fastmem_size):
                value = "MSM6242B"
            else:
                value = "none"
        elif c.uae_chipset_compatible == "A2000":
            value = "MSM6242B"
        elif c.uae_chipset_compatible == "A3000":
            value = "RP5C01A"
        elif c.uae_chipset_compatible == "A3000T":
            value = "RP5C01A"
        elif c.uae_chipset_compatible == "A4000":
            value = "RP5C01A"
        elif c.uae_chipset_compatible == "A4000T":
            value = "RP5C01A"
        else:
            f.fail("unknown uae_chipset_compatible")
            raise Exception("Failed")
    c.uae_rtc = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_sana2(c, f):
    if c.uae_sana2.explicit:
        # FIXME: ok? keep already specified value
        value = c.uae_rtc.explicit
        # FIXME: match and normalize uae boolean
    else:
        value = "false"
    c.uae_sana2 = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_slirp_implementation(c, f):
    if c.uae_slirp_implementation.explicit:
        # FIXME: ok? keep already specified value
        value = c.uae_slirp_implementation.explicit
        # FIXME: check value
    else:
        value = "auto"
    c.uae_slirp_implementation = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_toccata(c, f):
    if c.uae_toccata.explicit:
        value = c.uae_toccata.explicit
        if f.matches(value, ["true", "yes", "1"]):
            value = "true"
        elif f.matches(value, ["false", "no", "0"]):
            value = "false"
        else:
            f.warning(value + ": invalid value")
            value = "false"
    elif c.sound_card == "toccata":
        value = "true"
    else:
        value = "false"
    c.uae_toccata = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_z3chipmem_size(c, f):
    if c.uae_z3chipmem_size.explicit:
        value = c.uae_z3chipmem_size.explicit
    else:
        value = "0"
    c.uae_z3chipmem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uae_z3mem_size(c, f):
    if c.uae_z3mem_size.explicit:
        value = c.uae_z3mem_size.explicit
    elif c.zorro_iii_memory:
        value = str(int(c.zorro_iii_memory) // 1024)
    else:
        value = "0"
    c.uae_z3mem_size = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uaegfx_card(c, f):
    if c.uaegfx_card.explicit:
        # FIXME: check boolean
        value = c.uaegfx_card.explicit
    else:
        value = ""
    c.uaegfx_card = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _uaenative_library(c, f):
    if c.uaenative_library.explicit:
        value = c.uaenative_library.explicit
    else:
        value = "0"
    c.uaenative_library = value


# noinspection PyUnusedLocal,SpellCheckingInspection,PyUnresolvedReferences
def _zorro_iii_memory(c, f):
    if c.zorro_iii_memory.explicit:
        value = c.zorro_iii_memory.explicit
    else:
        value = "0"
    c.zorro_iii_memory = value


class AbstractExpandFunctions:
    @staticmethod
    def matches(a, b):
        pass

    @staticmethod
    def fail(message):
        pass

    @staticmethod
    def warning(message):
        pass

    @staticmethod
    def lower(s):
        pass


def expand_config(c, f):
    assert isinstance(f, AbstractExpandFunctions)
    _amiga_model(c, f)
    _accelerator(c, f)
    _accelerator_memory(c, f)
    _bsdsocket_library(c, f)
    _cdrom_drive_0(c, f)
    _cdrom_image_0(c, f)
    _int_model(c, f)
    _cdrom_drive_count(c, f)
    _chip_memory(c, f)
    _cpu(c, f)
    _dongle_type(c, f)
    _fast_memory(c, f)
    _floppy_drive_0(c, f)
    _floppy_drive_1(c, f)
    _floppy_drive_2(c, f)
    _floppy_drive_3(c, f)
    _platform(c, f)
    _floppy_drive_count(c, f)
    _floppy_drive_speed(c, f)
    _fpu(c, f)
    _freezer_cartridge(c, f)
    _graphics_card(c, f)
    _uae_cpu_24bit_addressing(c, f)
    _uae_gfxcard_type(c, f)
    _uae_gfxcard_size(c, f)
    _graphics_memory(c, f)
    _uae_cpuboard_type(c, f)
    _int_accelerator_name(c, f)
    _slow_memory(c, f)
    _int_bogomem_size(c, f)
    _int_chipmem_size(c, f)
    _uae_chipset(c, f)
    _int_chipset_name(c, f)
    _uae_cpu_model(c, f)
    _uae_fpu_model(c, f)
    _uae_mmu_model(c, f)
    _int_cpu_name(c, f)
    _int_cpuboardmem1_size(c, f)
    _int_default_floppy_type(c, f)
    _int_fastmem_size(c, f)
    _int_graphics_card_bus(c, f)
    _int_graphics_card_name(c, f)
    _int_kickstart_ext_sha1(c, f)
    _int_kickstart_sha1(c, f)
    _int_kickstart_revision(c, f)
    _int_kickstart_version(c, f)
    _motherboard_ram(c, f)
    _uae_a3000mem_size(c, f)
    _int_mbresmem_low_size(c, f)
    _int_model_name(c, f)
    _int_ppc_model(c, f)
    _zorro_iii_memory(c, f)
    _uae_z3mem_size(c, f)
    _int_z3fastmem_size(c, f)
    _uae_bsdsocket_emu(c, f)
    _uae_chipset_compatible(c, f)
    _uae_rtc(c, f)
    _uae_sana2(c, f)
    _uae_z3chipmem_size(c, f)
    _int_uae_boot_rom(c, f)
    _int_z3chipmem_size(c, f)
    _jit_compiler(c, f)
    _joystick_port_0_mode(c, f)
    _joystick_port_1_mode(c, f)
    _joystick_port_2_mode(c, f)
    _joystick_port_3_mode(c, f)
    _kickstart_file(c, f)
    _kickstart_ext_file(c, f)
    _network_card(c, f)
    _sound_card(c, f)
    _uae_a2065(c, f)
    _uae_bogomem_size(c, f)
    _uae_cd32cd(c, f)
    _uae_cd32fmv(c, f)
    _uae_chipmem_size(c, f)
    _uae_fastmem_size(c, f)
    _uae_floppy0(c, f)
    _uae_floppy0type(c, f)
    _uae_floppy1(c, f)
    _uae_floppy1type(c, f)
    _uae_floppy2(c, f)
    _uae_floppy2type(c, f)
    _uae_floppy3(c, f)
    _uae_floppy3type(c, f)
    _uae_joyport0(c, f)
    _uae_joyport0autofire(c, f)
    _uae_joyport1(c, f)
    _uae_joyport1autofire(c, f)
    _uaenative_library(c, f)
    _uae_native_code(c, f)
    _uae_ppc_model(c, f)
    _uae_slirp_implementation(c, f)
    _uae_toccata(c, f)
    _uaegfx_card(c, f)
